__author__ = "diarmuid"
import sys

sys.path.append("..")
import os

import unittest
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct
import logging

logging.basicConfig(level=logging.DEBUG)

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class SimpleEthernetTest(unittest.TestCase):
    ######################
    # Ethernet
    ######################
    def test_DefaultEthernet(self):
        e = SimpleEthernet.Ethernet()
        self.assertEqual(e.dstmac, 0)
        self.assertEqual(e.srcmac, 0)
        self.assertEqual(e.type, SimpleEthernet.Ethernet.TYPE_IP)
        self.assertEqual(e.payload, bytes())

    def test_basicEthernet(self):
        """Create an ethernet frame, then unpack it to a new object"""
        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = 0x998877665544
        e.type = SimpleEthernet.Ethernet.TYPE_IP
        e.payload = struct.pack("H", 0xA)
        ethbuf = e.pack()

        e2 = SimpleEthernet.Ethernet()
        e2.unpack(ethbuf)

        self.assertEqual(e2.dstmac, 0x998877665544)
        self.assertEqual(e2.type, SimpleEthernet.Ethernet.TYPE_IP)
        self.assertEqual(e2.srcmac, 0x001122334455)

    def test_buildEmptyEthernet(self):
        """Try and create an empty ethernet frame"""
        e = SimpleEthernet.Ethernet()
        self.assertTrue(e.pack())

    def test_non_def_type(self):
        e = SimpleEthernet.Ethernet()
        e.type = SimpleEthernet.Ethernet.TYPE_PAUSE
        e.dstmac = 0x0180C2000001
        e.srcmac = 0x1
        e.payload = struct.pack(">HH", 0x1, 0x2)
        e2 = SimpleEthernet.Ethernet()
        e2.unpack(e.pack())
        self.assertEqual(e, e2)
        self.assertEqual("SRCMAC=00:00:00:00:00:01 DSTMAC=01:80:C2:00:00:01 TYPE=0X8808", repr(e))

    def test_ethernet_fcs(self):
        e = SimpleEthernet.Ethernet()
        e.type = SimpleEthernet.Ethernet.TYPE_IP
        e.dstmac = 0x0180C2000001
        e.srcmac = 0x1
        e.payload = struct.pack(">HH", 0x1, 0x2)
        ex_fcs = struct.pack(">I", 0x6ED798BF)
        self.assertEqual(e.pack(fcs=True)[-4:], ex_fcs)
        e2 = SimpleEthernet.Ethernet()
        e2.unpack(e.pack(fcs=True), fcs=True)
        self.assertEqual(e, e2)

    def test_vlan(self):
        logging.getLogger("SimpleEthernet").setLevel(logging.DEBUG)
        e = SimpleEthernet.Ethernet()
        e.type = SimpleEthernet.Ethernet.TYPE_PAUSE
        e.vlan = True
        e.dstmac = 0x0180C2000001
        e.srcmac = 0x1
        e.payload = struct.pack(">HH", 0x1, 0x2)
        e2 = SimpleEthernet.Ethernet()
        e2.unpack(e.pack())
        self.assertEqual(e, e2)
        # p = pcap.Pcap("_vlan.pcap",mode='w')
        # r = pcap.PcapRecord()
        # r.setCurrentTime()
        # r.payload = e.pack()
        # p.write(r)
        # p.close()

    ######################
    # IP
    ######################
    def test_defaultIP(self):
        i = SimpleEthernet.IP()
        try:
            i.pack()
        except:
            self.assertTrue(True)
        else:
            self.assertTrue(False)

    def test_basicIP(self):
        i = SimpleEthernet.IP()
        i.dstip = "235.0.0.1"
        i.srcip = "192.168.1.1"
        i.payload = struct.pack(">H", 0xA5)
        ippayload = i.pack()

        i2 = SimpleEthernet.IP()
        i2.unpack(ippayload)
        self.assertEqual(i2.srcip, "192.168.1.1")
        self.assertEqual(i2.dstip, "235.0.0.1")
        self.assertEqual(i2.payload, struct.pack(">H", 0xA5))

    def test_unpackIPShort(self):
        i = SimpleEthernet.IP()
        dummypayload = struct.pack("H", 0xA5)
        try:
            i.unpack(dummypayload)
        except:
            self.assertTrue(True)
        else:
            self.assertTrue(False)

    ######################
    # UDP
    ######################

    def test_defaultUDP(self):
        u = SimpleEthernet.UDP()
        self.assertTrue(u.pack())

    def test_basicUDP(self):
        u = SimpleEthernet.UDP()
        u.dstport = 5500
        u.srcport = 4400
        u.payload = struct.pack("B", 0x5)
        mypacket = u.pack()
        self.assertEqual(mypacket, struct.pack(">HHHHB", 4400, 5500, 9, 0, 0x5))
        self.assertEqual(repr(u), "SRCPORT=4400 DSTPORT=5500")

    def test_unpackUDPShort(self):
        u = SimpleEthernet.UDP()
        dymmypayload = struct.pack("H", 0xA5)
        self.assertRaises(ValueError, lambda: u.unpack(dymmypayload))

    ######################
    # ICMP
    ######################

    def test_defICMP(self):
        i = SimpleEthernet.ICMP()
        self.assertEqual(i.pack(), b"\x00\x00\xff\xff\x00\x00\x00\x00")

    ######################
    # Read a complete pcap file
    ######################
    def test_readUDP(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "test_input.pcap"))
        mypcaprecord = p[0]
        p.close()
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        self.assertEqual(e.srcmac, 0x0018F8B84454)
        self.assertEqual(e.dstmac, 0xE0F847259336)
        self.assertEqual(e.type, 0x0800)
        self.assertEqual(repr(e), "SRCMAC=00:18:F8:B8:44:54 DSTMAC=E0:F8:47:25:93:36 TYPE=0X800")

        # checksum test
        (exp_checksum,) = struct.unpack_from("<H", e.payload, 10)
        ip_hdr_checksum = SimpleEthernet.ip_calc_checksum(e.payload[:10] + e.payload[12:20])
        self.assertEqual(exp_checksum, ip_hdr_checksum)
        i = SimpleEthernet.IP()
        i.unpack(e.payload)
        self.assertEqual(i.dstip, "192.168.1.110")
        self.assertEqual(i.srcip, "213.199.179.165")
        self.assertEqual(i.protocol, 0x6)
        self.assertEqual(i.ttl, 48)
        self.assertEqual(i.flags, 0x2)
        self.assertEqual(i.id, 0x4795)
        self.assertEqual(i.len, 56)
        self.assertEqual(i.version, 4)
        # print i
        self.assertEqual(repr(i), "SRCIP=213.199.179.165 DSTIP=192.168.1.110 PROTOCOL=TCP LEN=56")

    def test_ipv4fragment(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ipv4frags.pcap"))
        pw = pcap.Pcap(os.path.join(THIS_DIR, "combined.pcap"), mode="w")

        r = pcap.PcapRecord()
        e = SimpleEthernet.Ethernet()
        e.unpack(p[0].packet)
        i1 = SimpleEthernet.IP()
        i1.unpack(e.payload)
        e.unpack(p[1].packet)
        p.close()
        i2 = SimpleEthernet.IP()
        i2.unpack(e.payload)
        self.assertEqual(i1.id, i2.id)
        combined = SimpleEthernet.combine_ip_fragments([i1, i2])
        e.payload = combined.pack()
        self.assertEqual(1428, len(combined.pack()))
        r.payload = e.pack()
        pw.write(r)
        pw.close()

    # Write an ICMP
    def test_writeICMP(self):
        p = pcap.Pcap("_icmp.pcap", mode="w")
        r = pcap.PcapRecord()
        r.setCurrentTime()

        ping_req = SimpleEthernet.ICMP()
        ping_req.type = SimpleEthernet.ICMP.TYPE_REQUEST
        ping_req.code = 0
        ping_req.request_id = 0x100
        ping_req.request_sequence = 123
        ping_req.payload = struct.pack(">32B", *range(32))

        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = 0x998877665544
        e.type = SimpleEthernet.Ethernet.TYPE_IP

        i = SimpleEthernet.IP()
        i.dstip = "235.0.0.1"
        i.srcip = "192.168.1.1"
        i.protocol = SimpleEthernet.IP.PROTOCOLS["ICMP"]
        i.payload = ping_req.pack()
        e.payload = i.pack()
        r.packet = e.pack()
        p.write(r)
        ping_req.type = SimpleEthernet.ICMP.TYPE_REPLY
        i.payload = ping_req.pack()
        e.payload = i.pack()
        p.close()

        p = pcap.Pcap("_icmp.pcap", mode="w")
        r = pcap.PcapRecord()
        r.setCurrentTime()
        r.packet = e.pack()
        p.write(r)
        p.close()

    def test_readIPchecksum(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "inetx_test.pcap"))
        mypcaprecord = p[0]
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        i = SimpleEthernet.IP()
        self.assertTrue(i.unpack(e.payload))
        p.close()

    @unittest.skip("AFDX broken")
    def test_afdx(self):
        af = SimpleEthernet.AFDX()
        af.type = 1
        af.networkID = 2
        af.equipmentID = 3
        af.interfaceID = 4
        af.vlink = 5
        af.sequencenum = 0
        af.payload = struct.pack(">50B", *range(50))
        self.assertEqual(len(af.pack()), 65)
        b = af.pack()
        af2 = SimpleEthernet.AFDX()
        af2.unpack(b)
        self.assertTrue(af == af2)

    def test_igmp(self):
        # https://www.cloudshark.org/captures/b2e93fcea0c2
        exp_b = struct.pack(">HHHHHHHH", 0x2200, 0xE338, 0x0, 0x1, 0x400, 0x0, 0xEFC3, 0x702)
        act_b = SimpleEthernet.IGMPv3.join_groups(["239.195.7.2"])
        self.assertEqual(exp_b, act_b)
        exp_b = struct.pack(">6I", 0x2200F33C, 0x2, 0x2000000, 0xEFC30702, 0x2000000, 0xEFFFFFFA)
        act_b = SimpleEthernet.IGMPv3.join_groups(["239.195.7.2", "239.255.255.250"])
        self.assertEqual(exp_b, act_b)

    def test_writeIGMP(self):
        p = pcap.Pcap("_igmp.pcap", mode="w")
        r = pcap.PcapRecord()
        act_b = SimpleEthernet.IGMPv3.join_groups(["239.195.7.2", "239.255.255.250"])

        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = SimpleEthernet.IGMPv3.MAC_ADDR[SimpleEthernet.IGMPv3.IP_ADDR_JOIN]
        e.type = SimpleEthernet.Ethernet.TYPE_IP

        i = SimpleEthernet.IP()
        i.dstip = SimpleEthernet.IGMPv3.IP_ADDR_JOIN
        i.srcip = "192.168.1.1"
        i.protocol = SimpleEthernet.IP.PROTOCOLS["IGMP"]
        FCS_LEN = 4
        pad_len = 64 - (len(act_b) + SimpleEthernet.Ethernet.HEADERLEN + SimpleEthernet.IP.IP_HEADER_SIZE + FCS_LEN)
        i.payload = act_b + struct.pack(">{}B".format(pad_len), *([0] * pad_len))
        e.payload = i.pack()
        r.packet = e.pack(fcs=True)
        p.write(r)
        p.close()

    ######################
    # ARP
    ######################

    def test_ARP(self):
        a = SimpleEthernet.ARP()
        a.dstip = "192.168.28.2"
        a.dstmac = 0x0
        a.srcip = "192.168.28.5"
        a.srcmac = 0xAF72B6D5B
        buf = a.pack()
        b = SimpleEthernet.ARP()
        self.assertIsNone(b.unpack(buf))
        self.assertEqual(a, b)
        p = pcap.Pcap("_arp.pcap", mode="w")
        r = pcap.PcapRecord()
        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = 0xFF_FF_FF_FF_FF_FF
        e.type = SimpleEthernet.Ethernet.TYPE_ARP
        e.payload = buf
        r.packet = e.pack(fcs=False)
        p.write(r)
        p.close()


if __name__ == "__main__":
    unittest.main()



"""
.. module:: MPEGTS
    :platform: Unix, Windows
    :synopsis: Class to handle MPEG Transport Streams

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import sys
import typing
import logging

logger = logging.getLogger(__name__)


TSC = {0: "Not Scrambled", 1: "Reserved", 2: "Scrambled even key", 3: "Scrambled odd key"}
ADAPTION_CTRL = {1: "Payload Only", 2: "Adaption Only", 3: "Adaption and Payload", 0: "Reserved"}

PY3 = sys.version_info > (3,)


class MPEGAdaptionExtension(object):
    """
    The MPEG Adaptation extension format https://en.wikipedia.org/wiki/MPEG_transport_stream#Packet

    """

    def __init__(self) -> None:
        self.ltw_flag: bool = False
        self.piecewise_rate_flag: bool = False
        self.seamless_splice_flag: bool = False
        self.ltw = bytes()
        self.piecewise = bytes()
        self.seamless_splice = bytes()

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, MPEGAdaptionExtension):
            return False
        for attr in self.__dict__.keys():
            if getattr(self, attr) != getattr(__value, attr):
                return False
        return True

    def pack(self) -> bytes:
        if len(self.ltw) == 2:
            self.ltw_flag = True
        elif len(self.ltw) == 0:
            self.ltw_flag = False
        else:
            raise Exception("ltw should be 2 bytes")

        if len(self.piecewise) == 3:
            self.piecewise_rate_flag = True
        elif len(self.piecewise) == 0:
            self.piecewise_rate_flag = False
        else:
            raise Exception("piecewise should be 3 bytes")

        if len(self.seamless_splice) == 5:
            self.seamless_splice_flag = True
        elif len(self.seamless_splice) == 0:
            self.seamless_splice_flag = False
        else:
            raise Exception("seamless should be 5 bytes")

        _len = 2 + len(self.ltw) + len(self.piecewise) + len(self.seamless_splice)
        _flags = (
            0x1F
            + (int(self.ltw_flag) << 7)
            + (int(self.piecewise_rate_flag) << 6)
            + (int(self.seamless_splice_flag << 5))
        )
        return struct.pack(">BB", _len, _flags) + self.ltw + self.piecewise + self.seamless_splice

    def unpack(self, buffer: bytes) -> int:
        """Unpack the buffer into an MpegAdaptionExtension and return offset of data used

        Args:
            buffer (bytes): _description_

        Returns:
            bytes: _description_
        """
        _len, _flags = struct.unpack_from(">BB", buffer)
        if len(buffer) < _len:
            raise Exception(f"Indicated len={_len} bufferlen ={len(buffer)}")
        payload = buffer[:_len]
        self.ltw_flag = bool((_flags >> 7) & 0x1)
        self.piecewise_rate_flag = bool((_flags >> 6) & 0x1)
        self.seamless_splice_flag = bool((_flags >> 5) & 0x1)
        offset = 2
        if self.ltw_flag:
            self.ltw = payload[offset : offset + 2]
            offset += 2
        if self.piecewise_rate_flag:
            self.piecewise = payload[offset : offset + 3]
            offset += 3
        if self.seamless_splice_flag:
            self.seamless_splice = payload[offset : offset + 5]
            offset += 5
        return offset

    def __repr__(self) -> str:
        return f"ltw_flag={self.ltw_flag}, piecewise_flag={self.piecewise_rate_flag}, seamless_flag={self.seamless_splice_flag}"


class MPEGAdaption(object):
    """
    The MPEG Adaptation field format https://en.wikipedia.org/wiki/MPEG_transport_stream#Packet

    """

    def __init__(self) -> None:
        self.length: int = 0
        self.discontinutiy: bool = False
        self.random_access: bool = False
        self.es_priority: bool = False
        self.pcr_flag: bool = False
        self.opcr_flag: bool = False
        self.splicing_flag: bool = False
        self.transpart_flag: bool = False
        self.extension_flag: bool = False
        self.pcr = bytes()
        self.opcr = bytes()
        self.splice_countdown = 0
        self.private_data = bytes()
        self.adaption_extension: typing.Optional[MPEGAdaptionExtension] = None

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, MPEGAdaption):
            return False
        for attr in self.__dict__.keys():
            if getattr(self, attr) != getattr(__value, attr):
                return False
        return True

    def __repr__(self) -> str:
        _r = f"Discontunity={self.discontinutiy}, random={self.random_access} Elementary Stream Indicator={self.es_priority} PCR={self.pcr_flag} "
        _r += f"OPCR={self.opcr_flag} Splicing Point Flag={self.splicing_flag} Transport Private Data={self.transpart_flag}, Adaption Extension={self.extension_flag}"
        if self.adaption_extension is not None:
            _r += repr(self.adaption_extension)
        return _r

    def unpack(self, buffer: bytes):
        self.length, _flags = struct.unpack_from(">BB", buffer)
        self.discontinutiy = bool((_flags >> 7) & 1)
        self.random_access = bool((_flags >> 6) & 1)
        self.es_priority = bool((_flags >> 5) & 1)
        self.pcr_flag = bool((_flags >> 4) & 1)
        self.opcr_flag = bool((_flags >> 3) & 1)
        self.splicing_flag = bool((_flags >> 2) & 1)
        self.transpart_flag = bool((_flags >> 1) & 1)
        self.extension_flag = bool((_flags >> 0) & 1)

        offset = 2
        if self.pcr_flag:
            self.pcr = buffer[offset : offset + 6]
            offset += 6

        if self.opcr_flag:
            self.opcr = buffer[offset : offset + 6]
            offset += 6

        if self.splicing_flag:
            self.splice_countdown = struct.unpack_from(">B", buffer, offset)
            offset += 1

        if self.transpart_flag:
            (_transport_len,) = struct.unpack_from(">B", buffer, offset)
            offset += 1
            self.private_data = buffer[offset : offset + _transport_len]
            offset += _transport_len
        if self.extension_flag:
            self.adaption_extension = MPEGAdaptionExtension()
            try:
                offset += self.adaption_extension.unpack(buffer[offset:])
            except Exception as e:
                logger.error(f"Failed to unpack the MPEGAdationExtension. Err={e}")
        _stuffing = buffer[offset:]
        return None

    def pack(self):

        if len(self.pcr) > 0:
            if len(self.pcr) != 6:
                raise Exception(f"PCR length ={len(self.pcr)}")
            self.pcr_flag = True
        if len(self.opcr) > 0:
            self.opcr_flag = True
        if self.splice_countdown > 0:
            self.splicing_flag = True
            splice_buf = struct.pack(">B", self.splice_countdown)
        else:
            splice_buf = bytes()
        if len(self.private_data) > 0:
            _transport_len_b = struct.pack(">B", len(self.private_data))
            self.transpart_flag = True
        else:
            _transport_len_b = bytes()
        if self.adaption_extension is not None:
            self.extension_flag = True
            _extension_buffer = self.adaption_extension.pack()
        else:
            _extension_buffer = bytes()
        _data_len = (
            len(self.pcr) + len(self.opcr) + len(self.private_data) + len(_extension_buffer) + len(splice_buf) + 1
        )
        if self.length > _data_len:
            _suffing_len = self.length - _data_len
        else:
            _suffing_len = 0
            self.length = _data_len
        _stuffing = b"\xff" * _suffing_len
        _flags = (
            (int(self.discontinutiy) << 7)
            + (int(self.random_access) << 6)
            + (int(self.es_priority) << 5)
            + (int(self.pcr_flag) << 4)
            + (int(self.opcr_flag) << 3)
            + (int(self.splicing_flag << 2))
            + (int(self.transpart_flag << 1))
            + int(self.extension_flag)
        )

        return (
            struct.pack(">BB", self.length, _flags)
            + self.pcr
            + self.opcr
            + splice_buf
            + _transport_len_b
            + self.private_data
            + _extension_buffer
            + _stuffing
        )


ADAPTION_PAYLOAD_AND_ADAPTION = 0x3
ADAPTION_PAYLOAD_ONLY = 0x1
ADAPTION_ADAPTION_ONLY = 0x2


class MPEGPacket(object):
    """
    The MPEGPacket is the elementary unit in an MPEG Transport Stream
    It contains an header, in which there's a sync word, continuity counter, and a _payload
    """

    def __init__(self):
        self.sync: int = 0x47
        self.pid: int = 0
        self.tei: bool = False
        self.pusi: bool = True
        self.transport_priority: int = 0
        self.tsc: int = 0
        self.adaption_ctrl: int = 0
        self.continuitycounter: int = 0
        self.payload: bytes = bytes()
        self.adaption_field: typing.Optional[MPEGAdaption] = None

    def unpack(self, buf: bytes):
        """
        Converts a buffer into an MPEGTS packet

        :param buf: The buffer to unpack into an MPEG Packet
        :type buf: str
        :rtype: bool
        """
        (self.sync, pid_full, counter_full) = struct.unpack_from(">BHB", buf)
        if self.sync != 0x47:
            raise Exception(f"Sync word={self.sync:#0X} not 0x47")

        self.pid = pid_full & 0x1FFF
        self.tei = bool((pid_full >> 15) & 0x1)
        self.pusi = bool((pid_full >> 14) & 0x1)
        self.transport_priority = (pid_full >> 13) & 0x1

        self.continuitycounter = counter_full & 0xF
        self.adaption_ctrl = (counter_full >> 4) & 0x3
        self.tsc = (counter_full >> 6) & 0x3

        if self.adaption_ctrl == ADAPTION_PAYLOAD_AND_ADAPTION:  # payload + adaption
            (adaption_len,) = struct.unpack_from(">B", buf[4:])
            if adaption_len > 0:
                # logger.debug(f"Adaption len ={adaption_len}")
                self.payload = buf[(4 + 1 + adaption_len) :]
                self.adaption_field = MPEGAdaption()
                try:
                    self.adaption_field.unpack(buf[4 : (adaption_len + 1 + 4)])
                except Exception as e:
                    logger.error(f"Failed to unpack the MPEGAdationExtension. Err={e}")
            else:
                self.payload = buf[(4 + 1) :]
        elif self.adaption_ctrl == ADAPTION_ADAPTION_ONLY:
            self.adaption_field = MPEGAdaption()
            try:
                self.adaption_field.unpack(buf[4:])
            except Exception as e:
                logger.error(f"Failed to unpack the MPEGAdationExtension. Err={e}")
            self.payload = bytes()
        elif self.adaption_ctrl == ADAPTION_PAYLOAD_ONLY:
            self.payload = buf[4:]
        else:
            logger.debug("Adaption control of 0 is reserved")

    def pack(self, nostuff: bool = False) -> bytes:
        pid_full = self.pid + (self.transport_priority << 13) + (int(self.pusi) << 14) + (int(self.tei) << 15)
        continuity = self.continuitycounter + (self.adaption_ctrl << 4) + (self.tsc << 6)
        _payload = struct.pack(">BHB", self.sync, pid_full, continuity)
        if self.adaption_ctrl == ADAPTION_ADAPTION_ONLY or self.adaption_ctrl == ADAPTION_PAYLOAD_AND_ADAPTION:
            if self.adaption_field is None:
                _adaption_fieldn_paylad = struct.pack(">B", 0)
            else:
                _adaption_fieldn_paylad = self.adaption_field.pack()
        else:
            _adaption_fieldn_paylad = bytes()

        _unstuffed = _payload + _adaption_fieldn_paylad + self.payload
        _l = len(self.payload)
        if nostuff:
            _stuffing = bytes()
        else:
            _stuffing = b"\xff" * (188 - len(_unstuffed))
        return _unstuffed + _stuffing

    def __repr__(self) -> str:
        r = f"PID={self.pid:#0X} PUSI={self.pusi} TSC={TSC[self.tsc]} Adaption={ADAPTION_CTRL[self.adaption_ctrl]}"
        if self.adaption_field is not None:
            r += "\n Adaption=" + repr(self.adaption_field)
        return r

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, MPEGPacket):
            return False
        match_attr = [
            "sync",
            "pid",
            "transport_priority",
            "tei",
            "pusi",
            "continuitycounter",
            "tsc",
            "adaption_ctrl",
            "payload",
            "adaption_field",
        ]
        for attr in match_attr:
            if getattr(self, attr) != getattr(__value, attr):
                return False
        return True


class MPEGTS(object):
    """
    This class handles MPEG Transport Streams.
    https://en.wikipedia.org/wiki/MPEG_transport_stream

    Each transport stream contains 188 byte packets.
    These packets contain either video, audio or metadata information

    :type blocks: list[MPEGPacket]

    """

    def __init__(self):
        self.blocks: typing.List[MPEGPacket] = []  #: List of MPEGPacket objects

    def append(self, block: MPEGPacket):
        return self.blocks.append(block)

    def unpack(self, buf: bytes):
        """
        This method will convert a buffer of bytes into an array of MPEG TS packets

        :param buf: The buffer to unpack
        :type buf: str

        :rtype: bool
        """

        remainingbytes = 0
        while remainingbytes < len(buf):
            MpegBlock = MPEGPacket()
            try:
                MpegBlock.unpack(buf[remainingbytes : remainingbytes + 188])
            except Exception as e:
                raise Exception(f"Could not decom data as MPEG. Error={e} data offset={remainingbytes}")
            self.blocks.append(MpegBlock)
            remainingbytes += 188

        return True

    def pack(self) -> bytes:
        __bytes = bytes()
        for block in self:
            __bytes += block.pack()

        return __bytes

    def __len__(self):

        return len(self.blocks)

    def __iter__(self):
        self._index = 0
        return self

    def next(self) -> MPEGPacket:
        if self._index < len(self.blocks):
            _block = self.blocks[self._index]
            self._index += 1
            return _block
        else:
            raise StopIteration

    __next__ = next

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, MPEGTS):
            return False
        if len(self) != len(__value):
            return False
        for idx in range(len(self)):
            if self[idx] != __value[idx]:
                return False

        return True

    def __repr__(self) -> str:
        _r = ""
        for _block in self:
            _r += f"{repr(_block)}\n"
        return _r

    def __getitem__(self, _key):
        if _key >= len(self):
            raise IndexError
        return self.blocks[_key]

    def NumberOfBlocks(self):
        raise DeprecationWarning("This is being deprecated")

    def FirstCount(self):
        raise DeprecationWarning("This is being deprecated")

    def LastCount(self):
        raise DeprecationWarning("This is being deprecated")



# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# 0.7.4 - Correct class declaration to include object
# 0.7.5 - Added an alias for streamid to IEAN
# 0.7.6 - Added repr function to SimpleEthernet
# 0.8.0 - Lots of updates and documentation updates
# 0.8.1 - Added IENA-N and IENA-D packets
# 0.9.0 - Added NDP
# 0.9.2 - Added Ch10
# 0.9.4 - Previous chapter 10 was the UDP wrapper. Correct Chapter10 packet now added
# 0.9.5 - Added ARINC data packet
# 0.10.0 - Added iNET
# 0.10.1 - Added info to exception
# 0.10.2 - Fixed IP checksum bug
# 0.11.0 - Chapter10 endianness changes
# 0.11.1 - IENA-N and IENA-D throw exception if there are not integer number of D and N parameters in the payload
# 0.11.2 - ParserAligned throws exception if the number of quadbytes is illegal
# 0.12.0 - Added IENA-M and IENA-Q
# 0.12.1 - Changed pcap to an iterator model
# 0.12.2 - Fixed iNET byte order
# 0.12.3 - Fixed iNET byte order
# 0.12.4 - Changed INET to separate seconds and nanoseconds
# 0.13.0 - Added UART Payload for Chapter 10
# 0.13.1 - Bug fixes on UART for Chapter 10
# 0.13.2 - Added NPD Segments for RS232
# 0.13.3 - Parser aligned printout
# 0.14.0 - Made compatiable with python3
# 0.15.0 - Added ch7 and updated all unit test to pass in both py3 py2
# 0.15.1 - Added unpack to ParserAligned
# 0.15.2 - Fixd the ascii example
# 0.15.3 - Minor updates to Ch7 and unittest
# 0.15.4 - Pcap comment removal
# 0.15.5 - Handled padding for UART data words in chapter 10
# 0.15.6 - Optimisations on Golay encoding
# 0.15.7 - Minor updates
# 0.15.8 - Traffic generator update
# 0.15.10 - Traffic generator update
# 0.15.11 - Added FCS Support to ethernet packets
# 0.15.12 - Indentation bugfix
# 0.15.13 - Renamed PDFR to PTFR
# 0.15.14 - Massive perf improvements with Golay handling
# 0.15.15 - Added summary to the recorder
# 0.15.16 - Added summary to validation script
# 0.15.17 - Added IGMPv2 simplified packet generation
# 0.15.18 - Updated validate and pkt generation script
# 0.15.19 - Fixed divide by error in validate_pcap script
# 0.15.20 - Updated tx script to be much more accurate with timing
# 0.15.21 - Added configparsed file for contol
# 0.15.22 - Added timestamp to packets in validation script
# 0.15.23 - Fixed compatibility with python2 and reduced size of test input
# 0.15.24 - Added rec time check on validate script
# 0.15.25 - Fixed unpacking of vlan. Expanded SimpleEthernet to handle VLAN more completely
# 0.15.26 - Validate script update
# 0.15.27 - Added simple IENA packet generation script
# 0.15.28 - Fixed line endings
# 0.16.0  - Refactored Chapter10 module
# 0.16.1  - Initialized attributes
# 0.16.2  - Lots of chapter10 updates.
# 0.16.3  - Added support for PCM throughput mode
# 0.16.4  - Added custom NPD segment for PCM Packetizer (0x60)
# 0.16.5  - NDP clean up.
# 0.16.6  - NDP clean up #2
# 0.16.7  - Minor clean up
# 0.16.8  - Added pcap to ch10 conversion script
# 0.16.9  - Made script executable
# 0.16.10 - Fixed wrapping of TMATs
# 0.16.11 - Added attribute to ch10 to indicate presence of secondary header
# 0.16.12 - Added timeid argument to the adau script
# 0.16.13 - Fixed sequence number rollover in adau_to_ch10 conversion. Makde optimisations to the PCM sync word length
# 0.16.14 - use pinksheet RTC time conversion. Add verification to endianness swap. Swap endianness on ARINC payloads
# 0.16.15 - fixed bug on ARINC conversion
# 0.16.16 - fixed bug in PTP -> RTC conversion
# 0.16.17 - fixed python3.11 only code in adau conversion script
# 0.16.18 - added source distribution
# 0.16.19 - Fixed timestamp bug in adau conversion script
# 0.16.20 - Changed RTC conversion to use Decimal as integer resolution was not sufficient to convert the timestamp
# 0.16.21 - adau conversion script accepts a folder input
# 0.17.0  - significant change to the MPEGTS block. Added classes for the adaption fields
# 0.17.1  - Added Video to the chapter10
# 0.17.2  - Major update to the MPEGTS section. Split out code into multiple modules. Added support for PAT packets
# 0.17.3  - Added support for PES + STANAG4609 packets
# 0.17.4  - Option to mpeg ts pack to not stuff
# 0.17.5  - Check chapter10 data length field and throw exception if not correct
# 0.17.6  - Reverted last change
# 0.17.7  - Allowed ch10 UART packet to be little endian
# 0.17.8  - Fixed ch10 secondary header checksum calculation
# 0.17.9  - Downgraded the offset error in ch7 to a warning. Added type hintint to ch7
# 0.17.10 - Added support for ARP in SimpleEthernet
# 0.17.11 - Added the SamDec class to support capturing live data from a SamDec
# 0.17.13 - Added chapter 10 recorder script.
# 0.17.14 - Fixed the PMT packet in MPEGTS
# 0.17.15 - Added support for reading SamDec pcap files
# 0.17.16 - Added documentation details
# 0.17.17 - Docstring updates
# 1.0.0   - First 1.0 release. No functional change from 0.17.17
# 1.1.0   - Moved all Chapter10 stuff into IRIG106 and then into the module according to the spec. Existing code will work but with a Deprecation warning
# 1.1.1   - Version tag not updates
# 1.1.2   - Fixed the ch10 examples
# 1.1.4   - Updates to the MPEGTS packets to build and decom a PTS / DTS packet
# 1.1.5   - Missing ut file added
# 1.1.6   - No chnage but tagging as 1.1.6
# 1.1.7   - Added logging error for incorrect IP checksum
# 1.1.8   - No change but tagging as 1.1.8
# 1.1.9   - Fixed ARINC ch10 intra packet header
# 1.2.0   - Moved Chapter 7 into the IRIG106 folder. Removed old Chapter10 directory
# 1.2.1   - Fixed logging in IP error message
# 1.2.2   - Fixed chapter 7 packet generation
# 1.2.3   - Last release was nto successful
# 1.2.4   - Removed exit in the ch7
# 1.2.5   - Added C implementation of Golay. Also improved the existing python impl
# 1.2.6   - Further optimisation of the C Golay implementation
# 1.2.7   - Changed setup so as not to break if extension cannot be compiled
# 1.2.8   (FJP 2025-07-01)
#         - Added Context Manager support to Pcap.py
#         - Allow Golay.decode() to accept bytearray or other bytes-like object instead of only bytes (if not int)
#         - simplified Golay.py; removed some redundant checks
__version__ = "1.2.8"



import struct


class ADTS(object):
    def __init__(self) -> None:
        self.aac: bytes = bytes()
        self.version: int = 0
        self.sampling_freq: int = 0
        self._length: int = 0
        self.no_crc: bool = True

    def unpack(self, buffer):
        words = struct.unpack_from(">7B", buffer)
        sw = ((words[1] >> 4) << 8) + words[0]
        if sw != 0xFFF:
            raise Exception(f"Sync word = {sw:#0X}")
        self.sampling_freq = (words[2] >> 2) & 0xF
        self.no_crc = bool(words[1] & 0x1)
        self._length = (words[5] >> 5) + (words[4] << 3) + ((words[3] & 0x3) << 11)
        if self.no_crc:
            self.aac = buffer[7:]
        else:
            self.aac = buffer[9:]

    def __repr__(self) -> str:
        return f"ADTS: NoCRC={self.no_crc} SamplFreq={self.sampling_freq} len={self._length} lenaac={len(self.aac)}"



#!/usr/bin/env python

"""
# Very rudimentary (but fast) validation of recorded data
# Finds all inetx packets and validate no missing sequence numbers
# It will also take the url to an axn mem and verify data after downloading
"""


import AcraNetwork.Pcap as pcap
import glob
import os.path
import struct
import time
import logging
import argparse
import json
from urllib.parse import urlparse
from urllib.request import urlopen
from os import mkdir, remove
import datetime
from dataclasses import dataclass, field
import typing
import zstandard as zstd
import sys

VERSION = "0.5.0"

logging.basicConfig(level=logging.INFO, format="%(levelname)-6s %(asctime)-15s %(message)s")


@dataclass
class Streams:
    streamid: int
    sequence: int
    pkt_count: int
    start_ts: float
    end_ts: float
    dropcnt: int
    rstcnt: int
    length: int
    datavol: int
    sequence_list: typing.List[int] = field(default_factory=list)

    def pps(self) -> int:
        if self.end_ts - self.start_ts <= 0:
            return 0
        return int(self.pkt_count / (self.end_ts - self.start_ts))

    def bitrate(self) -> int:
        if self.end_ts - self.start_ts <= 0:
            return 0
        return int(self.length * 8 * self.pkt_count / (self.end_ts - self.start_ts))

    def timelen(self) -> float:
        return self.end_ts - self.start_ts

    def drops_to_hist(self):
        bins = 30
        start_seq = self.sequence - self.pkt_count - self.dropcnt
        bin_wdth = int((self.sequence - start_seq) / bins)
        bin_cnt = [0] * bins
        for i in range(bins):
            for s in self.sequence_list:
                # print(f"{i}:{start_seq}:{s}:{bin_wdth}")
                if (start_seq + bin_wdth * i) <= s < (start_seq + bin_wdth * (i + 1)):
                    bin_cnt[i] += 1
        rstring = "|"
        for cnt in bin_cnt:
            if cnt == 0:
                rstring += " "
            else:
                if cnt > 9:
                    cnt = "*"
                rstring += f"{cnt}"
        rstring += "|"
        return rstring


def create_parser():
    # Argument parser
    parser = argparse.ArgumentParser(description="Validate inetx sequence numbers quickly in a pcap file")
    # Common
    parser.add_argument(
        "--folder",
        type=str,
        required=True,
        default=None,
        help="folder to parser for pcap files. Can be a URL",
    )
    parser.add_argument(
        "--verbose",
        required=False,
        action="store_true",
        default=False,
        help="verbose mode",
    )
    parser.add_argument(
        "--summary",
        required=False,
        action="store_true",
        default=False,
        help="only print summaries per file",
    )
    parser.add_argument(
        "--control",
        type=int,
        required=False,
        default=0x11000000,
        help="control field value",
    )
    parser.add_argument(
        "--histogram",
        required=False,
        action="store_true",
        default=False,
        help="print a rough histogram of where the drops happened",
    )
    parser.add_argument("--version", action="version", version="%(prog)s {}".format(VERSION))

    return parser


# recorder data
# tcpdump -i eth0 -W 999 -C 1000 -n -K -s 2000 -w rec.pcap.


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False


def is_compressed(x: str) -> bool:
    (_root, _ext) = os.path.splitext(x)
    return _ext == ".zst"


def main(args):
    roll_over = pow(2, 32)
    fnames = {}

    if uri_validator(args.folder):
        is_url = True
        all_files = []
        with urlopen(args.folder) as response:
            response_content = response.read()
            json_resp = json.loads(response_content)
            for f, e in json_resp.items():
                all_files.append(e["url"])
                fnames[e["url"]] = f
    else:
        is_url = False
        # Find all the files and sort by extension
        all_files = glob.glob(os.path.join(args.folder, "*.pcap*"))
    all_files.sort()

    # For recording the data
    streams: typing.Dict[int, Streams] = {}
    inetx_pkts_validate = 0
    data_count_bytes = 0
    start_t = time.time()
    loss_count = 0
    loss_data = 0
    total_pkt_count = 0

    for pfile in all_files:
        # To calculate the rate per pcap
        first_pcap_time = None
        last_pcap_time = None
        packet_data_vol = 0
        tmp_folder = ".tmp"
        loss = 0
        floss = 0
        outf = ""

        if is_url:
            CHUNK = 32 * 1024
            if not os.path.exists(tmp_folder):
                mkdir(tmp_folder, 0o755)
            outf = os.path.join(tmp_folder, fnames[pfile])
            sd = time.time()
            with urlopen(pfile) as response, open(outf, "wb") as out_file:
                data_len = 0
                while True:
                    chunk = response.read(CHUNK)
                    if not chunk:
                        break
                    data_len += len(chunk)
                    out_file.write(chunk)

                dlspeed = data_len * 8 / (time.time() - sd) / 1e6
                logging.info("Downloaded {} at {:.1f}Mbps and wrote to {}".format(pfile, dlspeed, outf))
            p = pcap.Pcap(outf)
        elif is_compressed(pfile):

            if not os.path.exists(tmp_folder):
                mkdir(tmp_folder, 0o755)
            fname_pcap, _ext = os.path.splitext(os.path.basename(pfile))
            decompressed_f = os.path.join(tmp_folder, fname_pcap)
            with open(pfile, "rb") as compressed:
                with open(decompressed_f, "wb") as decompressed:
                    dctx = zstd.ZstdDecompressor()
                    dctx.copy_stream(compressed, decompressed)

            # Get sizes
            compressed_size = os.path.getsize(pfile)
            decompressed_size = os.path.getsize(decompressed_f)

            # Compute compression ratio
            compression_ratio = decompressed_size / compressed_size
            logging.info(f"Compression ratio: {compression_ratio:.2f}x")
            p = pcap.Pcap(decompressed_f)
        else:
            p = pcap.Pcap(pfile)
        prev_rec_ts = None
        for i, r in enumerate(p):

            if first_pcap_time is None:
                first_pcap_time = r.sec + r.usec * 1e-6

            last_pcap_time = r.sec + r.usec * 1e-6

            # Do a check on the record timestamp
            if prev_rec_ts is not None:
                if prev_rec_ts > last_pcap_time:
                    delta = prev_rec_ts - last_pcap_time
                    logging.warning(f"Record={i + 1} Record timestamp negative jump {delta}s")
            prev_rec_ts = last_pcap_time

            packet_data_vol += len(r.payload)
            total_pkt_count += 1
            if len(r.payload) >= (26 + 0x28):  # For short packets don't try to decode them as inetx
                # pull out the key fields
                (dst_port, udp_len, checksum, control, stream_id, seq, _len, ptpsec, ptpnsec) = struct.unpack_from(
                    ">HHHIIIIII", r.payload, 0x24
                )
                if control == args.control:
                    if stream_id in streams:
                        stream = streams[stream_id]
                        if seq != (stream.sequence + 1) % roll_over:
                            pkt_ts = datetime.datetime.fromtimestamp(r.sec + r.usec * 1e-6).strftime(
                                "%H:%M:%S.%f %d %b"
                            )
                            if seq < stream.sequence:
                                logging.warning(
                                    "Source Restarted. File={} PktNum={} StreamID={:#0X} PrevSeq={} "
                                    "CurSeq={}".format(
                                        pfile,
                                        i,
                                        stream_id,
                                        stream.sequence,
                                        seq,
                                    )
                                )
                                stream.rstcnt += 1
                            else:
                                loss = seq - ((stream.sequence + 1) % roll_over)
                                if not args.summary:
                                    logging.error(
                                        "File={} TS={} PktNum={} StreamID={:#0X} PrevSeq={} CurSeq={} Lost={} Lost={:,} bytes".format(
                                            pfile,
                                            pkt_ts,
                                            i,
                                            stream_id,
                                            stream.sequence,
                                            seq,
                                            loss,
                                            loss * stream.length,
                                        )
                                    )
                                loss_count += loss
                                loss_data += loss * stream.length
                                stream.dropcnt += loss
                                stream.sequence_list.append(stream.sequence + 1)
                                floss += loss
                        stream.sequence = seq
                        stream.pkt_count += 1
                        stream.end_ts = ptpsec + ptpnsec / 1e9
                        stream.datavol += len(r.payload)
                    else:
                        stream = Streams(
                            stream_id,
                            seq,
                            1,
                            ptpsec + ptpnsec / 1e9,
                            ptpsec + ptpnsec / 1e6,
                            0,
                            0,
                            len(r.payload),
                            len(r.payload),
                            [],
                        )
                        streams[stream_id] = stream
                    inetx_pkts_validate += 1
                    data_count_bytes += len(r.payload)
        p.close()
        # The data rate at which we are validating

        try:
            dr = (data_count_bytes * 8) / (1e6 * (time.time() - start_t))
        except:
            dr = 100

        try:
            ave_rec_rate_mbps = (packet_data_vol * 8) / (last_pcap_time - first_pcap_time) / 1e6
        except:
            ave_rec_rate_mbps = 0
        sids_found = len(streams)
        if first_pcap_time is not None:
            file_stamp = datetime.datetime.fromtimestamp(first_pcap_time).strftime("%H:%M:%S %d %b")
        else:
            file_stamp = "unknown"
        info_str = (
            f"In {os.path.basename(pfile)} starting at {file_stamp}, {inetx_pkts_validate:10} packets validated. "
            f"Total_data={data_count_bytes/1e6:8.0f}KB  Lost={floss:5} StreamsFound={sids_found:5} "
            f"RecordRate={ave_rec_rate_mbps:5.0f}Mbps ValRate={dr:5.0f}Mbps"
        )
        if loss > 0:
            logging.error(info_str)
        else:
            logging.info(info_str)
        if args.verbose:
            if len(streams) > 0:
                logging.info(
                    "{:>7s} {:>9s} {:>9s} {:>9s} {:>9s}".format("SID", "Seq", "LostCount", "ResetCnt", "Length")
                )
            for sid, stream in streams.items():
                logging.info(
                    "{:#07X} {:9d} {:9d} {:9d} {:9d}".format(
                        sid,
                        stream.sequence,
                        stream.dropcnt,
                        stream.rstcnt,
                        stream.length,
                    )
                )
        if is_url and loss == 0:
            remove(outf)
        elif is_url and not args.verbose:
            remove(outf)
        if is_compressed(pfile):
            remove(decompressed_f)

    print("\n")
    if len(streams) > 0:
        logging.info(
            "{:>7s} {:>15s} {:>9s} {:>9s} {:>9s} {:>9s} {:>9s} {:>18s} {:>12s} {:>12s} {:>12s}".format(
                "SID",
                "Cnt",
                "LostCount",
                "ResetCnt",
                "Length",
                "PPS",
                "Mbps",
                "Elapsed Time(s)",
                "DataVol(MB)",
                "DropVol(Bytes)",
                "BitRate(Mbps)",
            )
        )
    for sid, stream in sorted(streams.items()):
        if args.histogram:
            _hist = stream.drops_to_hist()
        else:
            _hist = ""
        logging.info(
            "{:#07X} {:15,d} {:9d} {:9d} {:9d} {:9d} {:9.1f} {:18.1f} {:12,.1f} {:12,d} {:12,.1f} {}".format(
                sid,
                stream.pkt_count,
                stream.dropcnt,
                stream.rstcnt,
                stream.length,
                stream.pps(),
                stream.bitrate() / 1e6,
                stream.timelen(),
                stream.datavol / 1e6,
                stream.dropcnt * stream.length,
                stream.datavol * 8 / (stream.timelen() * 1e6),
                _hist,
            )
        )

    print(
        f"\nSUMMARY:  RXPKTS={total_pkt_count:>15,}  RXINETX={inetx_pkts_validate:>15,}  "
        f"RXBYTES={data_count_bytes//1024:>15,.1f} KB LOSTPKTS={loss_count:>15,}  LOSTDATA={loss_data//1024:>15,.1f} KB"
    )


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    main(args)
    sys.exit(0)



from __future__ import annotations

"""
.. module:: Chapter7
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct Chapter7 Packets
    http://www.irig106.org/docs/106-17/chapter7.pdf

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2024"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"

import struct

from AcraNetwork.IRIG106.Chapter7 import Golay
import math
import logging
import typing
from enum import IntEnum


ch7_logger = logging.getLogger(__name__)


class PTDPContent(IntEnum):
    FILL = 0x0
    ASP = 0x1
    TEST_COUNTER = 0x2
    CHAPTER_10 = 0x3
    ETHERNET_MAC = 0x4
    IP = 0x5
    CHAPTER_24 = 0x6
    ILLEGAL = 0xF

    @classmethod
    def _missing_(cls, value):
        return PTDPContent.ILLEGAL


class PTDPFragment(IntEnum):
    COMPLETE = 0x0
    FIRST = 0x1
    MIDDLE = 0x2
    LAST = 0x3
    ILLEGAL = 0x4

    @classmethod
    def _missing_(cls, value):
        return PTDPFragment.ILLEGAL


PTDP_HDR_LEN = 0x6  # 24bits x2
PTFR_HDR_LEN = 0x4  # 1 byte unprotected and 3 bytes protected

PTDT_LLP_TRAILER_LEN = 1
PTDP_MAX_LEN = 0x800


class PTDPLengthError(Exception):
    pass


class PTDPRemainingData(Exception):
    pass


def _new_ptfr(ptfr_len: int, streamid: int = 0x1, golay=Golay.Golay()) -> PTFR:
    """Return a new PTFR object initialised with some useful values"""
    ptfr = PTFR(golay)
    ptfr.length = ptfr_len
    ptfr.streamid = streamid
    return ptfr


def datapkts_to_ptfr(
    eth_ch10_packets: typing.Iterable[bytes, bool], ptfr_len: int = 500, streamid: int = 0x1, golay=Golay.Golay()
) -> typing.Generator[PTFR, None, None]:
    """
    Generator that will take a generator for ethernet packet aligned payloads
    and return the PTFRs encapsulating the payload

    """
    # Create a new frame
    ptfr = _new_ptfr(ptfr_len, streamid, golay)
    for ptdp in datapkts_to_ptdp(eth_ch10_packets):

        remaining_space = ptfr.remaining_space()
        # logging.debug(f"Adding ptdp {ptdp} with remaining space {remaining_space}")
        # If we have a low latency packet but not enough space to insert it full then
        # insert a fill packet
        if ptdp.low_latency and remaining_space <= (len(ptdp) + PTDT_LLP_TRAILER_LEN):
            # logging.warning(f"Need to add fill packet to fill remaining space={remaining_space}")
            fill = ptdp_fill(remaining_space)
            remainder = ptfr.add_payload(fill.pack(), False)
            yield ptfr
            ptfr = _new_ptfr(ptfr_len, streamid, golay)
            if len(remainder) > 0:
                ptfr.add_payload(remainder)
                # logging.debug(f"Added remainder of len {len(remainder)}. Remainder={ptfr.remaining_space()}")

        # Add the packet and if there is remainder push out the ptfr packet and start a new one
        remainder = ptfr.add_payload(ptdp.pack(), ptdp.low_latency)
        ch7_logger.debug(f"PTDP ({ptdp}) to be added to PTRF leaving remainder of len={len(remainder)}")
        while not ptfr.has_space():
            # Spit out the full frame
            yield ptfr
            # Create anew frame
            ptfr = _new_ptfr(ptfr_len, streamid, golay)
            # Maybe the  PTFP is bigger than one frame, then split it across PTFRs.
            while len(remainder) > ptfr_len:
                ptfr.ptdp_offset = 0x3FF  # Signifies that we have all PTDP payload
                ptfr.add_payload(remainder[:ptfr_len])
                yield ptfr  # Frame is full
                remainder = remainder[ptfr_len:]  # If we have more data then save it for the next frame
                ptfr = _new_ptfr(ptfr_len, streamid, golay)

            if len(remainder) > 0:
                # Point to the first offset of the frame.
                ptfr.ptdp_offset = len(remainder)
                # Add the PTDS to the frame
                remainder = ptfr.add_payload(remainder)


def datapkts_to_ptdp(eth_ch10_packets: typing.Iterable[bytes, bool]) -> typing.Generator[PTDP, None, None]:
    """
    Generator that will take a generator for ethernet packet aligned payloads
    and return the PTDPs encapsulating the payload

    :param eth_ch10_packets:
    :type eth_ch10_packets: collections.Iterable[bytes, bool]
    :rtype: collections.Iterable[PTDP]
    """
    for buffer, llp in eth_ch10_packets:
        # The packet fits in one PTDP
        if len(buffer) <= PTDP_MAX_LEN:
            ptdp_pkt = PTDP()
            ptdp_pkt.low_latency = llp
            ptdp_pkt.fragment = PTDPFragment.COMPLETE
            ptdp_pkt.content = PTDPContent.ETHERNET_MAC
            ptdp_pkt.payload = buffer
            ptdp_pkt.length = len(buffer)
            yield ptdp_pkt
        else:
            # NEED to split the packet into multiple PTDP packets. How many?
            number_of_packets = int(math.ceil(float(len(buffer)) / PTDP_MAX_LEN))
            for i in range(number_of_packets):
                ptdp_pkt = PTDP()
                ptdp_pkt.low_latency = llp
                ptdp_pkt.content = PTDPContent.ETHERNET_MAC  # Assume it's Ethernet for the moment
                # Label the fragments
                if i == 0:
                    ptdp_pkt.fragment = PTDPFragment.FIRST
                    ptdp_pkt.payload = buffer[:PTDP_MAX_LEN]
                elif i == number_of_packets - 1:
                    ptdp_pkt.fragment = PTDPFragment.LAST
                    ptdp_pkt.payload = buffer[i * PTDP_MAX_LEN :]
                else:
                    ptdp_pkt.fragment = PTDPFragment.MIDDLE
                    ptdp_pkt.payload = buffer[PTDP_MAX_LEN * i : PTDP_MAX_LEN * (i + 1)]

                ptdp_pkt.length = len(ptdp_pkt.payload)
                yield ptdp_pkt


class PTDP(object):
    def __init__(self, golay=Golay.Golay()):
        self.payload: bytes = bytes()
        self.low_latency: bool = False
        self.length: int = 0
        self.content: PTDPContent = PTDPContent.FILL
        self.fragment: int = PTDPFragment.COMPLETE
        self._payload: bytes = bytes()
        self._golay: Golay.Golay = golay

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, val: bytes):
        if len(val) > PTDP_MAX_LEN:
            raise Exception(
                "One PTDP packet can only be of max length {}. Split your data into multiple PTDP packets".format(
                    PTDP_MAX_LEN
                )
            )

        self._payload = val
        self.length = len(val)

    def pack(self) -> bytes:
        """
        Convert a PTDP object into a string buffer

        :rtype: bytes
        """
        self.length = len(self.payload)
        msw = self.length & 0xFFF
        lsw = (self.length >> 12) + (self.fragment << 4) + (self.content << 6)

        return self._golay.encode(lsw, as_string=True) + self._golay.encode(msw, as_string=True) + self.payload

    def unpack(self, buffer: bytes) -> bytes:
        """
        Convert a buffer into a PTDP object returning the remaining buffer

        :type buffer: bytes
        :rtype: bytes
        """

        if len(buffer) < 6:
            raise PTDPRemainingData("Can't unpack less than the header length")

        lsw = self._golay.decode(buffer[:3])
        msw = self._golay.decode(buffer[3:6])

        self.length = msw + ((lsw & 0xF) << 12)
        self.fragment = PTDPFragment((lsw >> 4) & 0x3)
        self.content = PTDPContent((lsw >> 6) & 0xF)
        if self.length > PTDP_MAX_LEN:
            raise PTDPLengthError("GolayHdr=len={}. Must be corrupted".format(self.length))
        elif len(buffer[6:]) < self.length:
            raise PTDPRemainingData(
                "Not a full PTDP packet. Rest likely in next packet . Buffer length={} GolayHdr=len={} fragment={} content={}".format(
                    len(buffer[6:]), self.length, self.fragment, self.content
                )
            )
        self.payload = buffer[6 : self.length + 6]

        return buffer[self.length + 6 :]

    def __len__(self):
        return len(self.payload) + PTDP_HDR_LEN

    def __eq__(self, other: PTDP):
        if not isinstance(other, PTDP):
            return False
        for attr in ["payload", "length", "fragment", "content"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other: PTDP):
        return not self.__eq__(other)

    def __repr__(self):
        return f"PTDP: Len={self.length} Content={repr(self.content)} Fragment={repr(self.fragment)} LowLatency={self.low_latency}"


def ptdp_fill(total_len_min: int) -> PTDP:
    _p = PTDP()
    _p.content = PTDPContent.FILL
    if total_len_min < (PTDP_HDR_LEN + 1):
        payload_len = 1
    else:
        payload_len = total_len_min - PTDP_HDR_LEN
    _p.payload = struct.pack(f">{payload_len}B", *([0xFF] * payload_len))
    return _p


class PTFR(object):
    """
    Object to represent the PTFR frame
    Pass in a Golay object as creation of one is expensive so sharing and caching speeds things up a lot
    """

    def __init__(self, golay=Golay.Golay()):
        self.version: int = 0x0
        self.streamid: int = 0x0
        self.llp: bool = False
        self.ptdp_offset: int = 0x0
        self.length: int = 0
        self._payload: bytes = bytes()
        self._golay: Golay.Golay = golay

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, val: bytes):
        if len(val) + len(self._payload) > self.length:
            raise Exception("Length of payload ({}) is larger than length field ({})".format(len(val), self.length))
        self._payload += val

    def add_payload(self, buffer: bytes, is_llp: bool = False) -> bytes:
        """
        Add the buffer to the payload ensuring not to go over the length field
        Return the remainder
        """
        if len(buffer) == 0:
            raise Exception("Can't add payload of len = 0")
        if is_llp and len(buffer) + len(self._payload) > self.length:
            raise Exception(
                f"LLP packet ({len(buffer)}) cannot be added (ptfrlen={self.length}) as it will push the payload ({len(self.payload)}) into the next frame. Add a fill word"
            )
        if is_llp and len(self._payload) > 0 and self.llp:
            ch7_logger.debug(
                f"Adding an LLP buffer({len(buffer)}) to some existing llp payload ({len(self._payload)}). Shifting original data and adding llp at "
                "start. len={}".format(len(buffer) + 1)
            )
            self.ptdp_offset += len(buffer) + 1
            self._payload = buffer + struct.pack(">B", 0xFF) + self._payload
            self.llp = True
        elif is_llp and len(self._payload) > 0 and not self.llp:
            ch7_logger.debug(
                "Adding an LLP buffer to a payload with high latency data. Shifting the data and adding the "
                "offset. len={}".format(len(buffer) + 1)
            )
            self.ptdp_offset = len(buffer) + 1
            self._payload = buffer + struct.pack(">B", 0x0) + self._payload
            self.llp = True
        elif is_llp and len(self.payload) == 0:
            ch7_logger.debug("Adding first LLP buffer len={}".format(len(buffer) + 1))
            self.llp = True
            self._payload = buffer + struct.pack(">B", 0x00)  # Further LLPs will be added in front of this one
            self.ptdp_offset += len(buffer) + 1
        else:
            self._payload += buffer
            ch7_logger.debug(f"Adding a normal PTDP buffer of len={len(buffer)} to a total len={len(self._payload)}")

        if len(self._payload) > self.length:
            len_to_take = self.length - len(self._payload)
            remainder = self.payload[len_to_take:]
            self._payload = self.payload[:len_to_take]
            return remainder
        else:
            return bytes()

    def has_space(self, expected_addition: int = 0) -> bool:
        return (len(self.payload) + expected_addition) < self.length

    def remaining_space(self) -> int:
        return self.length - len(self.payload)

    def pack(self) -> bytes:
        """
        Convert a PTFR object into a string for transmission. This will return the packed string and the remainder string
        for any partial ptdp packet

        :param previous_partial: If we have some partial data from the previous PTFR packet, add it here
        :rtype: (bytes, bytes)
        """
        if len(self.payload) != self.length:
            raise Exception("Payload length does not match the length field")

        buffer = struct.pack(">B", self.version + (self.streamid << 4))
        buffer += self._golay.encode(self.ptdp_offset + (self.llp << 11), as_string=True)
        buffer += self.payload

        return buffer

    def unpack(self, buffer: bytes) -> bool:
        """
        Convert the PTFR data from one minor frame into a PTFR object
        :param buffer:
        :return:
        """
        (byte_,) = struct.unpack_from(">B", buffer)
        self.version = byte_ & 0x3
        self.streamid = (byte_ >> 4) & 0xF
        # Protected field
        protected_field = self._golay.decode(buffer[1:4])
        self.llp = bool((protected_field >> 11) & 0x1)
        self.ptdp_offset = protected_field & 0x7FF
        self.payload = buffer[4:]

        return True

    def check_offsets(self, act_offset: int) -> bool:
        if (act_offset != self.ptdp_offset) and (self.ptdp_offset != 2047):
            ch7_logger.debug(
                f"Offset of unpacked PTDP packet ({act_offset}) does not match the declared offset ({self.ptdp_offset})"
            )
            return False
        elif self.ptdp_offset != 2047:
            ch7_logger.debug(
                "Offset of unpacked PTDP packet ({}) matches the declared offset ({})".format(
                    act_offset, self.ptdp_offset
                )
            )
        return True

    def get_aligned_payload(
        self, first_PTFR: bool, remainder: typing.Optional[bytes] = None
    ) -> typing.Generator[typing.Tuple[PTDP, bytes, str], None, None]:
        """
        Return the payload as PTDP packets with the final partial payload
        The remainder is the bytes from the end of the previous PTFR. IF this is the middle of a
        capture set it to None so that false positive messages about offsets is triggered

        :type remainder: bytes
        :param remainder: Optional partial payload from previous frame
        :rtype: Tuple[PTDP, bytes, str]
        """
        aligned = True
        # The PTFR decides what is low latency initially
        is_llp = self.llp

        if is_llp:
            # ch7_logger.debug("LLP flag set. First packet should be LLP")
            buf = self.payload
        elif remainder is None and self.ptdp_offset > 0 and self.ptdp_offset < 0x7FF:
            buf = self.payload[self.ptdp_offset :]
            # ch7_logger.debug(
            #    "Start of analysis. Could be in the middle of a packet offset={} buffer length={}".format(
            #        self.ptdp_offset, len(buf)
            #    )
            # )
        elif remainder == bytes() and self.ptdp_offset > 0 and self.ptdp_offset < 0x7FF:
            buf = self.payload[self.ptdp_offset :]
            # ch7_logger.debug(
            #    "No remainder from previous packet, offset={} buffer length={}".format(self.ptdp_offset, len(buf))
            # )
        elif remainder is None:
            buf = self.payload
            # ch7_logger.debug(
            #    "Buffer length={}. Ignoring offset={} Remainder undefined".format(len(buf), self.ptdp_offset)
            # )
        else:
            buf = remainder + self.payload
            # ch7_logger.debug(
            #    "Buffer length={}. Ignoring offset={} Remainder length={}".format(
            #        len(buf), self.ptdp_offset, len(remainder)
            #    )
            # )

        do_offset_check = True
        if is_llp:
            byte_offset = 0
        elif remainder is None:  # Fake the offset if we have jumped into the middle of a data stream
            byte_offset = self.ptdp_offset
        else:
            byte_offset = -1 * len(remainder)
            if len(remainder) == 0:
                do_offset_check = False

        offset_check_count = 0

        while aligned:
            # ch7_logger.debug(f"Starting to check buf of lenght={len(buf)}")
            p = PTDP(self._golay)
            try:
                buf = p.unpack(buf)
            except PTDPLengthError as e:
                aligned = False
                ch7_logger.warning(
                    "Looks like we got an illegal PTDP length. Resetting. {}bytes. Message={} Offset={}".format(
                        len(buf), e, self.ptdp_offset
                    )
                )

                yield (None, None, e)
            except PTDPRemainingData as e:
                aligned = False
                # ch7_logger.debug(
                #    "Failed to unpack buffer of length {}bytes. Message={} Offset={}".format(
                #        len(buf), e, self.ptdp_offset
                #    )
                # )
                # if not is_llp and byte_offset > 0 and do_offset_check:
                #    self.check_offsets(byte_offset)

                yield (None, buf, e)
            else:
                if not is_llp and do_offset_check and byte_offset >= 0:
                    # ch7_logger.debug(f"do_offset_check={do_offset_check} byte_offset={byte_offset}")
                    # self.check_offsets(byte_offset)
                    do_offset_check = False
                    offset_check_count += 1
                elif not is_llp and not do_offset_check and offset_check_count < 1:
                    do_offset_check = True
                    byte_offset += len(p)
                    # ch7_logger.debug(
                    #    f"Enable do_offset_check because {offset_check_count} < 1 byte_offset={byte_offset} len_p={len(p)}"
                    # )
                elif not is_llp:
                    byte_offset += len(p)

                # set the low latency flag on the current packet now we know if we are at the end of the LLP sequence.
                p.low_latency = is_llp

                if is_llp:  # If this is a low latency packet
                    # Remove the last byte
                    (next_llp,) = struct.unpack_from(">B", buf)
                    # Check if the next PTDP is low latency before yielding
                    if next_llp == 0xFF:
                        # ch7_logger.debug("Next packet is LLP")
                        is_llp = True
                        buf = buf[1:]
                        byte_offset += len(p) + 1
                    else:
                        is_llp = False
                        # if ((remainder == bytes()) or first_PTFR)  and self.ptdp_offset > 0:
                        if ((remainder == bytes()) and self.ptdp_offset > 0) or first_PTFR:
                            # ch7_logger.debug("LLP Packets extracted, jumping to offset")
                            buf = self.payload[self.ptdp_offset :]
                            do_offset_check = False
                            byte_offset = self.ptdp_offset
                            offset_check_count = 1
                        elif remainder is None:
                            buf = buf[1:]
                            byte_offset += len(p) + 1
                        else:
                            buf = (
                                remainder + buf[1:]
                            )  # The remainder is only added after all the llp packets are removed
                            byte_offset += len(p) + 1 - len(remainder)
                            if len(remainder) > 0:
                                do_offset_check = False

                # ch7_logger.debug(f"Returning p={repr(p)} and no remainder")
                yield (p, bytes(), "")

        # ch7_logger.debug("------PTFR expired-----")

    def __eq__(self, other: PTFR):
        if not isinstance(other, PTFR):
            return False

        for attr in ["version", "streamid", "llp", "ptdp_offset", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return (
            f"PTFR: Length={self.length} StreamID={self.streamid:#0X} Offset={self.ptdp_offset} LowLatency={self.llp}\n"
        )



import unittest
import AcraNetwork.iNET as iNET
import os
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct
from base64 import b64encode

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def getEthernetPacket(data=""):
    e = SimpleEthernet.Ethernet()
    e.srcmac = 0x001122334455
    e.dstmac = 0x998877665544
    e.type = SimpleEthernet.Ethernet.TYPE_IP
    i = SimpleEthernet.IP()
    i.dstip = "235.0.0.1"
    i.srcip = "192.168.1.1"
    i.protocol = SimpleEthernet.IP.PROTOCOLS["UDP"]
    u = SimpleEthernet.UDP()
    u.dstport = 6678
    u.srcport = 6678
    u.payload = data
    i.payload = u.pack()
    e.payload = i.pack()
    return e.pack()


class testINET(unittest.TestCase):
    def setUp(self):
        self.pkg = iNET.iNETPackage()
        self.pkg.definitionID = 7
        self.pkg.flags = 8
        self.pkg.timedelta = 99
        self.pkg.payload = struct.pack(">B", 0xA)

        self.i = iNET.iNET()
        self.i.flags = 0
        self.i.type = 1
        self.i.definition_ID = 0xDC
        self.i.sequence = 3
        self.i.ptptimeseconds = 100
        self.i.ptptimenanoseconds = 1000
        self.i.app_fields = [0x30, 0x40]
        self.i.packages.append(self.pkg)

        for pkg_idx in range(2, 5):
            pkg = iNET.iNETPackage()
            pkg.definitionID = 7
            pkg.flags = 8
            pkg.timedelta = 0x99
            payload = list(range(pkg_idx))
            pkg.payload = struct.pack(">{}B".format(pkg_idx), *payload)
            self.i.packages.append(pkg)

    def test_unpack_compare(self):
        d = iNET.iNET()
        d.unpack(self.i.pack())
        self.assertEqual(d.flags, self.i.flags)
        self.assertEqual(d.type, self.i.type)
        self.assertEqual(d.type, self.i.type)
        self.assertEqual(d.definition_ID, self.i.definition_ID)
        self.assertEqual(d.sequence, self.i.sequence)
        self.assertEqual(d.ptptimeseconds, self.i.ptptimeseconds)
        self.assertEqual(d.ptptimenanoseconds, self.i.ptptimenanoseconds)
        self.assertEqual(d.app_fields, self.i.app_fields)
        self.assertEqual(d._payload, self.i._payload)
        for idx, pkg in enumerate(d.packages):
            self.assertEqual(pkg.payload, self.i.packages[idx].payload)
            self.assertEqual(pkg.definitionID, self.i.packages[idx].definitionID)
            self.assertEqual(pkg.timedelta, self.i.packages[idx].timedelta)
            self.assertEqual(pkg.flags, self.i.packages[idx].flags)

        self.assertTrue(self.i == d)

    def test_topcap(self):
        pcapw = pcap.Pcap(os.path.join(THIS_DIR, "test_inet.pcap"), mode="w")

        rec = pcap.PcapRecord()
        for pkg_idx in range(2, 5):
            pkg = iNET.iNETPackage()
            pkg.definitionID = 7
            pkg.flags = 8
            pkg.timedelta = 99
            payload = list(range(pkg_idx))
            pkg.payload = struct.pack(">{}B".format(pkg_idx), *payload)
            self.i.packages.append(pkg)
        rec.payload = getEthernetPacket(self.i.pack())
        pcapw.write(rec)
        pcapw.close()

    def test_print(self):
        self.assertEqual(
            repr(self.i),
            "MessageDefinitionID=0XDC Sequence=3 Type=1 TimeStamp(s)=100 TimeStamp(ns)=1000 OptionWordCount=0",
        )

    def test_len(self):
        self.assertEqual(len(self.i), 96)

    def test_pack(self):
        ref_str = self.i.pack()
        i2 = iNET.iNET()
        i2.unpack(ref_str)

        self.assertEqual(self.i, i2)

    def test_minimalinet(self):
        i = iNET.iNET()
        i.definition_ID = 6
        print(b64encode(i.pack()))


if __name__ == "__main__":
    unittest.main()



import socket
import logging
import AcraNetwork.iNetX as inetx
import AcraNetwork.Pcap as pcap
import typing
import struct


PCM_HDR_LEN = 10
logger = logging.getLogger(__name__)


def string_matching_boyer_moore_horspool(text: str = "", pattern: str = "") -> typing.List[int]:
    """
    Returns positions where pattern is found in text.
    O(n)
    Performance: ord() is slow so we shouldn't use it here
    Example: text = 'ababbababa', pattern = 'aba'
         string_matching_boyer_moore_horspool(text, pattern) returns [0, 5, 7]
    :param text: text to search inside
    :param pattern: string to search for
    :return: list containing offsets (shifts) where pattern is found inside text
    """
    m = len(pattern)
    n = len(text)
    offsets = []
    if m > n:
        return offsets
    skip = []
    for k in range(256):
        skip.append(m)
    for k in range(m - 1):
        my = pattern[k]
        skip[pattern[k]] = m - k - 1

    skip = tuple(skip)
    k = m - 1
    while k < n:
        j = m - 1
        i = k
        while j >= 0 and text[i] == pattern[j]:
            j -= 1
            i -= 1
        if j == -1:
            offsets.append(i + 1)
        k += skip[text[k]]

    return offsets


class SamDec008(object):
    """
    The SAM/DEC/008 is a USB power PCM decommutator (https://www.curtisswrightds.com/products/flight-test/ground-stations/samdec008)

    Once configured it will convert PCM frames into iNetX packets over UDP

    This class will capture UDP packets from the network, extract the iNetX payload and align the data to PCM frame
    boundaries. It will return PCM frames as bytes

    Supply the UDP port and the IP Address of the correct network interface card on your PC
    You can use ''  to let you OS decide

    :param udp_port: The receive UDP port number
    :type udp_port: int
    :param timeout: Timeout in seconds
    :type timeout: float
    :param localaddress: The local network interface ip address
    :type localaddress: str

    >>> samdec = SamDec008(8010, localaddress="127.0.0.1", timeout=0.5)
    >>> for frame in samdec.frames():
    ...     (syncword, sfid, word1) = struct.unpack_from(">IHH", frame)


    """

    def __init__(self, udp_port: int, timeout: float = 5.0, localaddress=""):

        self.recv_sockets = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.recv_sockets.settimeout(timeout)
        self.recv_sockets.bind((localaddress, udp_port))

        self._current_aligned = True
        self._payload_offset = 0
        self._sequence = None

        self.streamid = 0x153  #: StreamID on which to capture the SAM/DEC data. Default of 0x153 should be ok
        self.frame_length = None  #: This will be populated when seraching for frame sync words
        self.sync_word = 0xFE6B2840  #: The Frame sync word.

    def close(self):
        self.recv_sockets.close()

    def _get_data(self) -> typing.Generator[bytes, None, None]:
        """
        Get the data, agnostic to network vs pcap

        :rtype: collections.Iterable[str]
        """
        while True:
            try:
                data, addr = self.recv_sockets.recvfrom(10000)
            except Exception as e:
                yield None
            else:
                yield data

    def frames(self) -> typing.Generator[bytes, None, None]:
        """Get the data from the underlying source, combine the IP fragments and then pull out the payload from the
        inetx packets and align them

        Yields:
            bytes: the payload captured
        """
        sync_packed = struct.pack(">I", self.sync_word)
        for udp_payload in self._get_data():
            if udp_payload is None:
                return
            inetx_pkt = inetx.iNetX()
            try:
                inetx_pkt.unpack(udp_payload)
            except Exception as e:
                continue
            else:
                self._payload_offset = PCM_HDR_LEN  # The SAM DEC inserts some header
                if inetx_pkt.streamid == self.streamid:
                    if self._sequence is not None:
                        if self._sequence + 1 % pow(2, 64) == inetx_pkt.sequence:
                            logger.warning(
                                "Missing Sequence number at {}. Is SAM/DEC dropping?".format(inetx_pkt.sequence)
                            )
                    # Start with the previous segment and the current payload
                    payload = inetx_pkt.payload

                    if self.frame_length is None:
                        offset = string_matching_boyer_moore_horspool(payload, sync_packed)
                        if len(offset) == 0:
                            raise Exception("No Frame sync found")
                        elif len(offset) == 1:
                            self.frame_length = len(payload) - PCM_HDR_LEN
                        else:
                            self.frame_length = offset[1] - offset[0]

                    # If we get to within the last 4 bytes then hold it until the next segment
                    while (self._payload_offset + self.frame_length) <= len(payload):
                        # We have enough for a full frame
                        frame_buffer = payload[self._payload_offset : self._payload_offset + self.frame_length]
                        self._payload_offset += self.frame_length
                        # Verify that we are still in sync
                        if frame_buffer[:4] != sync_packed:
                            # Not in sync.
                            print("ERROR:Fell out of alignment at offset {}".format(self._payload_offset))
                            self.frame_length = None
                        else:
                            yield frame_buffer


class SamDecPcap(SamDec008):
    """
    The SAM/DEC/008 is a USB power PCM decommutator (https://www.curtisswrightds.com/products/flight-test/ground-stations/samdec008)

    Once configured it will convert PCM frames into iNetX packets over UDP

    This class will take iNetx packet fromn a pcap file, extract the iNetX payload and align the data to PCM frame
    boundaries. It will return PCM frames as bytes

    :param pcap_fname: The PCAP filename
    :type pcap_fname: str


    >>> samdec = SamDecPcap("test/sample_pcap/samdec.pcap")
    >>> for frame in samdec.frames():
    ...     (syncword, sfid, word1) = struct.unpack_from(">IHH", frame)
    ...     print(f"SW={syncword:#0X} sfid={sfid}")
    SW=0XFE6B2840 sfid=4
    SW=0XFE6B2840 sfid=5
    SW=0XFE6B2840 sfid=6
    SW=0XFE6B2840 sfid=7

    """

    def __init__(self, pcap_fname: str):

        self._pcap = pcap.Pcap(pcap_fname, mode="r")

        self._current_aligned = True
        self._payload_offset = 0
        self._sequence = None

        self.streamid = 0x153  #: StreamID on which to capture the SAM/DEC data. Default of 0x153 should be ok
        self.frame_length = None  #: This will be populated when seraching for frame sync words
        self.sync_word = 0xFE6B2840  #: The Frame sync word.

    def close(self):
        self._pcap.close()

    def _get_data(self) -> typing.Generator[bytes, None, None]:
        """
        Get the data, agnostic to network vs pcap

        :rtype: collections.Iterable[str]
        """
        UDP_TYPE = 17
        for rec in self._pcap:
            if len(rec.payload) > 0x46:
                (ip_type,) = struct.unpack_from(">B", rec.payload, 0x17)
                if ip_type == UDP_TYPE:
                    data = rec.payload[0x2A:]
                    logger.debug(f"FrameLen={len(data)}")
                    yield data



# -*- coding: utf-8 -*-
import datetime


class timedelta(datetime.timedelta):
    def __new__(cls, *args, **kwargs):
        nanoseconds = 0
        if 'nanoseconds' in kwargs.keys():
            nanoseconds = kwargs['nanoseconds']
            del kwargs['nanoseconds']
        if not 'microseconds' in kwargs.keys():
            kwargs['microseconds'] = 0

        if nanoseconds < 0:
            kwargs['microseconds'] -= 1

        kwargs['microseconds'] += int(nanoseconds / 1000.0)
        nanoseconds = nanoseconds % 1000

        c = super(timedelta, cls).__new__(cls, *args, **kwargs)
        c.nanoseconds = nanoseconds

        return c

    @property
    def _hour(self):
        return int(self.seconds / 3600)

    @property
    def _minute(self):
        return int(self.seconds / 60) % 60

    @property
    def _second(self):
        return self.seconds % 60

    def __str__(self, *args, **kwargs):
        days = ""
        micros = ""
        nanos = ""
        if not self.days == 0:
            pural = ""
            if abs(self.days) > 1:
                pural = "s"
            days = "{0:d} day{1}, ".format(self.days, pural)
        if not self.microseconds == 0 or not self.nanoseconds == 0:
            micros = ".{0:06d}".format(self.microseconds)
        if not self.nanoseconds == 0:
            nanos = ":{0:03d}".format(self.nanoseconds)
        return "{0}{1:d}:{2:02d}:{3:02d}{4}{5}".format(
            days,
            self._hour, self._minute, self._second,
            micros, nanos
        )


class nanotime(datetime.datetime):
    def __new__(cls, *args, **kwargs):
        nanosecond = 0
        if len(args) >= 8:
            if isinstance(args[7], int):
                nanosecond = args[7]
                args = args[0:7] + args[8:]
        if 'nanosecond' in kwargs.keys():
            nanosecond = kwargs['nanosecond']
            del kwargs['nanosecond']

        c = super(nanotime, cls).__new__(cls, *args, **kwargs)
        c.nanosecond = nanosecond
        return c

    def nanotimeClass(self, c):
        return nanotime(year=c.year,
                        month=c.month,
                        day=c.day,
                        hour=c.hour,
                        minute=c.minute,
                        second=c.second,
                        microsecond=c.microsecond,
                        nanosecond=c.nanosecond,
                        )

    def __str__(self):
        return "{0}-{1}-{2} {3:02d}:{4:02d}:{5:02d}.{6:06d}:{7:03d}".format(
            self.year, self.month, self.day,
            self.hour, self.minute, self.second,
            self.microsecond, self.nanosecond,
        )

    def __add__(self, *args, **kwargs):
        try:
            nanoseconds = args[0].nanoseconds
        except AttributeError:
            nanoseconds = 0

        if hasattr(args[0], 'days') and hasattr(args[0], 'seconds'):
            pass
        else:
            raise TypeError("unsupported operand type(s) for +: '{0}' and '{1}'".format(self.__class__.__name__,
                                                                                        args[0].__class__.__name__))

        c = super(nanotime, self).__add__(*args, **kwargs)

        nanoseconds += self.nanosecond
        microsecond = c.microsecond + int(nanoseconds / 1000.0)
        nanoseconds %= 1000
        # return self.nanotimeClass(c)

        return nanotime(c.year,
                        c.month,
                        c.day,
                        c.hour,
                        c.minute,
                        c.second,
                        microsecond,
                        nanoseconds,
                        )

    def __sub__(self, *args, **kwargs):
        if hasattr(args[0], 'days') and hasattr(args[0], 'seconds'):
            days = args[0].days
            seconds = args[0].seconds
            microseconds = args[0].microseconds
            try:
                nanoseconds = args[0].nanoseconds
            except AttributeError:
                nanoseconds = 0
            x = timedelta(days=-days,
                          seconds=-seconds,
                          microseconds=-microseconds,
                          nanoseconds=-nanoseconds,
                          )
            args = [x]
        elif hasattr(args[0], 'day') and hasattr(args[0], 'second'):
            t0 = self
            t1 = args[0]
            d0 = datetime.datetime(t0.year, t0.month, t0.day, t0.hour, t0.minute, t0.second)
            d1 = datetime.datetime(t1.year, t1.month, t1.day, t1.hour, t1.minute, t1.second)
            total_seconds = (d0 - d1).total_seconds()
            attr_list = ('year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond', 'nanosecond',)
            b = {}
            for a in attr_list:
                try:
                    b[a] = getattr(self, a) - getattr(args[0], a)
                except AttributeError:
                    b[a] = getattr(self, a)
            return timedelta(seconds=total_seconds,
                             microseconds=b['microsecond'],
                             nanoseconds=b['nanosecond'],
                             )
        else:
            raise TypeError("unsupported operand type(s) for +: '{0}' and '{1}'".format(self.__class__.__name__,
                                                                                        args[0].__class__.__name__))

        return self.__add__(*args, **kwargs)

    def replace(self, *args, **kwargs):
        attr_list = ['year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond', 'nanosecond']
        new_dict = {}
        for attr in attr_list:
            new_dict[attr] = getattr(self, attr)
        for kw in kwargs:
            if not kw in attr_list:
                raise NameError("Unknown attr: {0}".format(kw))
            new_dict[kw] = kwargs[kw]
        if new_dict['nanosecond'] >= 1000:
            raise ValueError("nanosecond must be in 0..999")
        return nanotime(new_dict['year'],
                        new_dict['month'],
                        new_dict['day'],
                        new_dict['hour'],
                        new_dict['minute'],
                        new_dict['second'],
                        new_dict['microsecond'],
                        new_dict['nanosecond'],
                        )

    @property
    def total_seconds(self):
        seconds = int((self - nanotime.utcfromtimestamp(0)).total_seconds())
        return seconds



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
"""
=====
 Sending IENA packets
=====

Simplified IENA packet generator. Sends one fixed size packet at the specified rate
Can be used as a base for generating more complex configurations

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2023"
__version__ = "0.1.0"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import sys
sys.path.append("..")
import time
import struct
import AcraNetwork.SimpleEthernet as eth
import AcraNetwork.IENA as iena
import argparse
import socket
import signal
import datetime


def create_parser():
    # Argument parser
    parser = argparse.ArgumentParser(description='Send IENA packets at a specified rate')
    parser.add_argument('--rate', required=False, type=float, default=1.0, help="Packet rate in Mbps")
    parser.add_argument('--ipaddress', required=False, type=str, default="192.168.0.26", help="Destination IP")

    return parser


def accurate_sleep(duration, get_now=time.perf_counter):
    now = get_now()
    end = now + duration
    while now < end:
        now = get_now()


def main(args):
    dst_udp_port = 4444

    # open a UDP socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

    # Get the accurate size of the packet
    FCS_LEN = 4
    PREAMBLE = 8
    hdr_lens = eth.UDP.UDP_HEADER_SIZE + eth.IP.IP_HEADER_SIZE + eth.Ethernet.HEADERLEN + FCS_LEN + PREAMBLE

    # Work out the usecond timestamp
    now = datetime.datetime.now()
    seconds_since_jan1st = (now - now.replace(day=1, month=1, hour=0, minute=0, second=0, microsecond=0)).total_seconds()
    useconds_since_jan1st = int(seconds_since_jan1st * 1e6)

    # Create an inetx packet
    myiena = iena.IENA()
    myiena.key = 0xDC
    myiena.keystatus = 0x0
    myiena.status = 0x0
    myiena.timeusec = useconds_since_jan1st
    myiena.payload = struct.pack(">700H", *(range(700)))
    myiena.sequence = 0

    # Figure how much of a gap between packets
    _payload_len = len(myiena.pack())
    total_vol_data = 0
    chunk_count_ps = args.rate * 1e6/(_payload_len * 8)
    tx_time_vol = total_vol_data * 8 * 1e-9
    gap_per_pkt = (1 - tx_time_vol) / (chunk_count_ps)
    if gap_per_pkt <= 0:
        gap_per_pkt = 0

    print("UDP target IP:", args.ipaddress)
    print("UDP target port:", dst_udp_port)
    print("Rate = {} Mbps".format(args.rate))
    print("DLY = {:.6f} s".format(gap_per_pkt))

    # keep a track of the packets sent
    sequence_roll_over = pow(2, 16)
    packet_count = 1
    vol_data_sent = 0

    # Handle the user interruption gracefully
    def signal_handler(*args):
        print(f"Exiting. Sent {packet_count:>12,} packets and {vol_data_sent:>15,} bytes")
        sys.exit()
    signal.signal(signal.SIGINT, signal_handler)

    # loop forever
    while True:
        mypayload  = myiena.pack()
        sock.sendto(mypayload, (args.ipaddress, dst_udp_port))
        vol_data_sent += (len(mypayload) + hdr_lens)
        # Increment the sequence number
        myiena.sequence = (myiena.sequence + 1) % sequence_roll_over
        # Add to the timeuseconds
        myiena.timeusec += int(gap_per_pkt * 1e6)
        # Sleep
        accurate_sleep(gap_per_pkt)
        # Information
        packet_count += 1
        if packet_count % 100 == 0:
            print(f"{packet_count:>12,} packets sent. {vol_data_sent:>18,} bytes send.")


if __name__ == '__main__':
    parser = create_parser()
    pargs = parser.parse_args()
    main(pargs)
    sys.exit(0)




# -------------------------------------------------------------------------------
# Name:
# Purpose:
#
# Author:
#
# Created:
#
__author__ = "$USER"

import sys

sys.path.append("..")

import unittest
import datetime
import time
import AcraNetwork.IENA as iena
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct
import os
from copy import copy
from base64 import b64encode

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def get_udp_packet(pcapfilename):
    """
    Convienence method to read the first UDP packet in a pcap file

    :param pcapfilename: pcapfilename
    :type pcapfilename: str
    :rtype: SimpleEthernet.UDP
    """
    p = pcap.Pcap(os.path.join(THIS_DIR, pcapfilename))
    p._read_global_header()
    mypcaprecord = p[0]
    e = SimpleEthernet.Ethernet()
    e.unpack(mypcaprecord.packet)
    ip = SimpleEthernet.IP()
    ip.unpack(e.payload)
    udp_pkt = SimpleEthernet.UDP()
    udp_pkt.unpack(ip.payload)
    p.close()

    return udp_pkt


ienq_q = """IENAQ: KEY=0XDC SEQ=2 TIMEUS=0 NUM_QPARAM=5
 Q-Param #0:ParamID=0XA Dataset Length=15
 Q-Param #1:ParamID=0XB Dataset Length=16
 Q-Param #2:ParamID=0XC Dataset Length=17
 Q-Param #3:ParamID=0XD Dataset Length=18
 Q-Param #4:ParamID=0XE Dataset Length=19
"""


class IENATest(unittest.TestCase):
    def setUp(self):
        # Get aUDP packet from a pack which can be used
        p = pcap.Pcap(os.path.join(THIS_DIR, "iena_test.pcap"))
        p._read_global_header()

        # Loop through the pcap file reading one packet at a time
        try:
            mypcaprecord = p[0]
        except IOError:
            # End of file reached
            return
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip = SimpleEthernet.IP()
        ip.unpack(e.payload)
        self.upd = SimpleEthernet.UDP()
        self.upd.unpack(ip.payload)
        p.close()

    def test_defaultIENA(self):
        """Check the defaults for the IENA packet"""
        i = iena.IENA()
        self.assertEqual(i.key, 0)
        self.assertEqual(i.size, 0)
        self.assertEqual(i.keystatus, 0)
        self.assertEqual(i.status, 0)
        self.assertEqual(i.sequence, 0)
        self.assertEqual(i.endfield, 0xDEAD)
        self.assertEqual(i.payload, bytes())

    def test_basicIENA(self):
        """Build a very basic packet and check the packed result"""
        i = iena.IENA()
        i.key = 1
        i.keystatus = 2
        i.status = 3
        i.sequence = 10
        i.setPacketTime(time.mktime(datetime.datetime(datetime.datetime.today().year, 1, 2, 0, 0, 0, 0).timetuple()), 0)
        i.payload = struct.pack("H", 0x5)
        # size = 9. Time = midnight Jan 02 = 86400s 0us = 0x14dd760000
        expected_payload = (
            struct.pack(iena.IENA.IENA_HEADER_FORMAT, 1, 9, 0x14, 0x1DD76000, 2, 3, 10)
            + struct.pack("H", 0x5)
            + struct.pack(">H", 0xDEAD)
        )
        self.assertEqual(i.pack(), expected_payload)

    def test_basicIENA_streamid(self):
        """Build a very basic packet and check the packed result"""
        i = iena.IENA()
        i.streamid = 1
        i.keystatus = 2
        i.status = 3
        i.sequence = 10
        self.assertEqual(i.key, i.streamid)
        self.assertEqual(i.status, i.n2)
        i.setPacketTime(time.mktime(datetime.datetime(datetime.datetime.today().year, 1, 2, 0, 0, 0, 0).timetuple()), 0)
        i.payload = struct.pack("H", 0x5)
        # size = 9. Time = midnight Jan 02 = 86400s 0us = 0x14dd760000
        expected_payload = (
            struct.pack(iena.IENA.IENA_HEADER_FORMAT, 1, 9, 0x14, 0x1DD76000, 2, 3, 10)
            + struct.pack("H", 0x5)
            + struct.pack(">H", 0xDEAD)
        )
        self.assertEqual(i.pack(), expected_payload)
        self.assertEqual(repr(i), "IENAP: KEY=0X1 SEQ=10 TIMEUS=86400000000")

    def test_unpackIEANFromPcap(self):
        """Read all the IENA packets in a pcap file and check each field"""

        # Now I have a _payload that will be an iena packet
        i = iena.IENA()
        i.unpack(self.upd.payload)
        self.assertEqual(i.key, 0x1A)
        self.assertEqual(i.size, 24)
        self.assertEqual(i.status, 1)
        self.assertEqual(i.keystatus, 1)
        self.assertEqual(i.sequence, 195)
        self.assertEqual(i.timeusec, 0x1D102F800)
        self.assertEqual(i.endfield, 0xDEAD)
        self.assertEqual(len(i), 48)
        i2 = copy(i)
        self.assertTrue(i2 == i)

    def test_unpack_IENAM(self):
        """Read all the IENA packets in a pcap file and check each field"""
        # Now I have a _payload that will be an iena packet
        i = iena.IENAM()
        i.unpack(self.upd.payload)
        self.assertEqual(i.key, 0x1A)
        for mparam in i:
            self.assertEqual(mparam.paramid, 0xDC)
            self.assertEqual(repr(mparam), "ParamID=0XDC Delay=16 Dataset Length=26")
        # print(i)
        self.assertEqual(
            repr(i),
            "IENAM: KEY=0X1A SEQ=195 TIMEUS=7801600000 NUM_MPARAM=1\n M-Param #0:ParamID=0XDC Delay=16 Dataset Length=26\n",
        )

    def test_unpack_IENAN(self):
        # Decode as IENA-N
        i = iena.IENAN()
        i.unpack(self.upd.payload)
        self.assertEqual(i.key, 0x1A)
        param_1 = i.parameters[0]  # type: iena.NParameter
        self.assertEqual(param_1.paramid, 0xDC)
        self.assertListEqual(param_1.dwords, [0x10])
        self.assertEqual(repr(i), "IENAN: KEY=0X1A SEQ=195 TIMEUS=7801600000 NUM_DPARAM=8")
        self.assertEqual(len(i), 8)
        for idx, n in enumerate(i):
            if idx == 0:
                self.assertEqual(n.paramid, 220)

    def test_unpack_IENAD(self):
        # Decode as IENA-N
        udp_pkt = get_udp_packet("ienad.pcap")
        i = iena.IENAD()
        i.unpack(udp_pkt.payload)
        self.assertEqual(i.key, 0x2CFA)
        param_1 = i.parameters[0]  # type: iena.DParameter
        self.assertEqual(param_1.paramid, 0xFFFF)
        self.assertEqual(param_1.delay, 0x0)
        self.assertListEqual(param_1.dwords, [0xFED1, 0x7CFE])
        self.assertEqual(repr(i), "IENAD: KEY=0X2CFA SEQ=0 TIMEUS=1837 NUM_DPARAM=11")
        self.assertEqual(len(i), 11)
        for id, p in enumerate(i):
            if id == 0:
                self.assertEqual(p.paramid, 0xFFFF)
        self.assertEqual(i[0].paramid, 0xFFFF)

    def test_corrupt_IENAD(self):
        udp_pkt = get_udp_packet("corrupt_ienad.pcap")
        iena_pkt = iena.IENAD()
        self.assertRaises(Exception, lambda: iena_pkt.unpack(udp_pkt.payload))

    def test_create_ienam(self):
        i = iena.IENAM()
        i.key = 0xDC
        i.endfield = 0xDEAD
        i.keystatus = 0
        i.sequence = 2
        i.n2 = 0
        for idx in range(5):
            mparam = iena.MParameter(paramid=idx, delay=idx * 2, dataset=os.urandom(idx + 5))
            i.parameters.append(mparam)

        buf = i.pack()
        b = iena.IENAM()
        b.unpack(buf)
        self.assertEqual(b.sequence, 2)
        self.assertEqual(len(b), 5)
        for idx, p in enumerate(b):
            self.assertEqual(p.paramid, idx)
            self.assertEqual(p.delay, idx * 2)
            self.assertEqual(len(p.dataset), idx + 5)

    def test_create_ienam_small(self):
        i = iena.IENAM()
        i.key = 0xDC
        i.endfield = 0xDEAD
        i.keystatus = 0
        i.sequence = 2
        i.n2 = 0
        for idx in range(1, 2):
            mparam = iena.MParameter(paramid=idx, delay=idx * 2, dataset=bytes(1))
            i.parameters.append(mparam)

        buf = i.pack()
        print(b64encode(buf))

    def test_create_ienaq_small(self):
        i = iena.IENAQ()
        i.key = 0xDC
        i.endfield = 0xDEAD
        i.keystatus = 0
        i.sequence = 2
        i.n2 = 0
        for idx in range(2, 3):
            mparam = iena.QParameter(paramid=idx, dataset=bytes(1))
            i.parameters.append(mparam)

        buf = i.pack()
        print(b64encode(buf))

    def test_create_ienaq(self):
        i = iena.IENAQ()
        i.key = 0xDC
        i.endfield = 0xDEAD
        i.keystatus = 0
        i.sequence = 2
        i.n2 = 0

        for idx in range(10, 15):
            qparam = iena.QParameter(paramid=idx, dataset=struct.pack(">{}B".format(idx + 5), *range(idx + 5)))
            if idx == 10:
                # print repr(qparam)
                self.assertEqual(repr(qparam), "ParamID=0XA Dataset Length=15")

            i.parameters.append(qparam)
        self.assertEqual(repr(i), ienq_q)
        buf = i.pack()
        b = iena.IENAQ()
        b.unpack(buf)
        self.assertEqual(b.sequence, 2)
        self.assertEqual(len(b), 5)
        for idx, p in enumerate(b):
            self.assertEqual(p.paramid, idx + 10)
            self.assertEqual(len(p.dataset), idx + 5 + 10)
        # print(i)

    def test_read_multiple_ienam(self):
        # Get aUDP packet from a pack which can be used
        p = pcap.Pcap(os.path.join(THIS_DIR, "ienam_multiple.pcap"))

        # Loop through the pcap file reading one packet at a time
        mypcaprecord = p[0]
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip = SimpleEthernet.IP()
        ip.unpack(e.payload)
        udp = SimpleEthernet.UDP()
        udp.unpack(ip.payload)

        i = iena.IENAM()
        i.unpack(udp.payload)
        self.assertEqual(len(i), 2)
        p.close()


if __name__ == "__main__":
    unittest.main()



__author__ = "diarmuid"
import unittest
from AcraNetwork import endianness_swap
import struct
import timeit


class MiscTest(unittest.TestCase):
    def test_endianness(self):
        buf = struct.pack(">BBBB", 1, 2, 3, 4)
        swap2 = struct.pack(">BBBB", 2, 1, 4, 3)
        swap4 = struct.pack(">BBBB", 4, 3, 2, 1)
        self.assertEqual(endianness_swap(buf), swap2)
        self.assertEqual(endianness_swap(buf, 4), swap4)

    def test_endianness_wrong_size(self):
        buf = struct.pack(">BBB", 1, 2, 3)
        try:
            endianness_swap(buf, 4)
        except:
            self.assertTrue(True)
        else:
            self.assertTrue(False)

    @unittest.skip("Performance check")
    def test_endiannessperf(self):
        rt = timeit.timeit(
            "endianness_swap(b'\x01\x02\x03\x04', 2)",
            number=1_000_000,
            setup="from AcraNetwork import endianness_swap",
        )
        #



"""
.. module:: Golay
    :platform: Unix, Windows
    :synopsis: Encode and decode Golay Codes based on IRIG106 Standard http://www.irig106.org/docs/106-17/chapter7.pdf

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2020"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


# This is a direct porting of the C code from the IRIG106 starndard.
import struct
from functools import lru_cache
import warnings

try:
    from . import golay_c as _golay_native

    _use_c_extension = True
except ImportError:
    warnings.warn("C extension for Golay not found. Falling back to pure Python.", RuntimeWarning)
    _use_c_extension = False


GOLAY_SIZE = 0x1000
G_P = [0xC75, 0x63B, 0xF68, 0x7B4, 0x3DA, 0xD99, 0x6CD, 0x367, 0xDC6, 0xA97, 0x93E, 0x8EB]
H_P = [0xA4F, 0xF68, 0x7B4, 0x3DA, 0x1ED, 0xAB9, 0xF13, 0xDC6, 0x6E3, 0x93E, 0x49F, 0xC75]


class Golay:
    """
    Encode and Decode Golay numbers
    """

    # Look-up tables are class variables. They will be initialised by the
    # constructor of the first instance of the Golay class to be created.
    # Set them to None here so the constructor can tell that they must be
    # set up.
    EncodeTable = None
    SyndromeTable = None
    CorrectTable = None
    ErrorTable = None

    def __init__(self):
        if _use_c_extension:
            _golay_native.golay_init_tables()
        else:
            if Golay.EncodeTable is None:        
                self._init_encode_table()
            self._initgolaydecode()


    def _init_encode_table(self):
        Golay.EncodeTable = [0] * GOLAY_SIZE
        for x in range(GOLAY_SIZE):
            Golay.EncodeTable[x] = x << 12
            for i in range(12):
                if x >> (11 - i) & 1:
                    Golay.EncodeTable[x] ^= G_P[i]

    def encode(self, raw, as_string=False):
        if not (0 <= raw <= 0xFFF):
            raise ValueError("Only 12-bit unsigned values allowed")

        if _use_c_extension:
            encoded = _golay_native.golay_encode(raw)
        else:
            encoded = self._encode_python(raw)

        if as_string:
            return encoded.to_bytes(3, "big")
        return encoded

    def decode(self, encoded):
        if _use_c_extension:
            return _golay_native.golay_decode(encoded)
        else:
            # encoded is either an integer <= 0xFFFFFF, or is a bytes-like type
            if not isinstance(encoded, int):
                if len(encoded) != 3:
                    raise ValueError("3-byte input required")
                v = int.from_bytes(encoded, "big")
            elif not (0 <= encoded <= 0xFFFFFF):
                raise ValueError("Only 24-bit unsigned values supported")
            else:
                v = encoded
            return self._decode_python(v)


    def _encode_python(self, raw):
        """
        Encode the value as a 24b code

        The leading '_' indicates this is a private method; do not call this
        directly, call encode() instead.

        :type raw: int
        :param raw: value to be encoded that is already validated to be 0..FFF
        :return: encoded value as a 24-bit integer
        """
        # self.encode() has already checked that 0 <= raw <= 0xFFF so do not 
        # check again
        # Also, there is no to_string argument because that is handled by
        # encode()

        return Golay.EncodeTable[raw & 0xFFF]

    def _decode_python(self, v):
        """
        Decode a 24b number as a golay

        The leading '_' indicates this is a private method; do not call this
        directly, call decode() instead.

        :type v: int
        :param v: integer that has already been validated to be 24bit
        :return: decoded 12-bit value
        """
        # self.decode() has converted the value to an integer and verified 
        # that it is valid. So do not repeat the check.
        
        return self._decode2(((v) >> 12) & 0xFFF, (v) & 0xFFF)

    def _syndrome2(self, v1, v2):
        return Golay.SyndromeTable[v2] ^ (v1)

    def _syndrome(self, v):
        return self._syndrome2(((v) >> 12) & 0xFFF, (v) & 0xFFF)

    def _errors2(self, v1, v2):
        return Golay.ErrorTable[self._syndrome2(v1, v2)]

    def _decode2(self, v1, v2):
        return (v1) ^ Golay.CorrectTable[self._syndrome2(v1, v2)]

    def _errors(self, v):
        return self._errors2(((v) >> 12) & 0xFFF, (v) & 0xFFF)

    def errors(self, v):
        if _use_c_extension:
            return _golay_native.golay_errors(v)
        else:
            return self._errors(v)

    @staticmethod
    def _onesincode(code, size):
        """Optimised version of the code below. Runs 2x"""
        return bin(code)[2 : size + 2].count("1")

    @staticmethod
    def _onesincode_old(code, size):
        ret = 0

        for t in range(size):
            if (code >> t) & 1:
                ret += 1

        return ret

    def _initgolaydecode(self):
        Golay.SyndromeTable = [0] * GOLAY_SIZE
        Golay.CorrectTable = [0] * GOLAY_SIZE
        Golay.ErrorTable = [0] * GOLAY_SIZE

        for x in range(GOLAY_SIZE):
            Golay.SyndromeTable[x] = 0
            for i in range(12):
                if (x >> (11 - i)) & 1:
                    Golay.SyndromeTable[x] ^= H_P[i]
                    Golay.ErrorTable[x] = 4
                    Golay.CorrectTable[x] = 0xFFF

        Golay.ErrorTable[0] = 0
        Golay.CorrectTable[0] = 0
        for i in range(24):
            for j in range(24):
                for k in range(24):
                    error = (1 << i) | (1 << j) | (1 << k)
                    syndrome = self._syndrome(error)
                    Golay.CorrectTable[syndrome] = (error >> 12) & 0xFFF
                    Golay.ErrorTable[syndrome] = Golay._onesincode(error, 24)

        return True



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
"""
=====
 Sending UDP packets
=====

Send UDP packets at a specific rate
This can be customised by providing an ini file. Format:

[inetxpayloadlength]
min=1300
max=1400
[randomisation]
#distribution=uniform
distribution=beta
alpha=3
beta=1
[tweaks]
packetbuildtime=0.0000070


"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__version__ = "0.4.0"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import sys
sys.path.append("..")
import time
import struct
import AcraNetwork.iNetX as inetx
import AcraNetwork.SimpleEthernet as eth
import argparse
import socket
import random
import signal
import configparser
from collections import namedtuple
from dataclasses import dataclass


ConfigSetting = namedtuple("ConfigSetting", ['min', 'max', 'dist', 'alpha', 'beta', 'buildtime', 'accuratets'])

NANOSECS = int(1e9)


@dataclass
class TS:
    sec: int
    nsec: int

    def add(self, gap):
        ns = int(gap * 1e9)
        _dlt = self.nsec + ns
        self.nsec = (_dlt % NANOSECS)
        self.sec += int(_dlt/NANOSECS)

    def pack(self):
        return struct.pack(">II", self.sec, self.nsec)


def get_default_settings(configfile: str) -> ConfigSetting:
    config = configparser.ConfigParser()
    config.read(configfile)
    min = int(config.get('inetxpayloadlength', 'min', fallback='64'))
    max = int(config.get('inetxpayloadlength', 'max', fallback='1400'))
    dist = (config.get('randomisation', 'distribution', fallback='uniform'))
    alpha = float(config.get('randomisation', 'alpha', fallback='3.0'))
    beta = float(config.get('randomisation', 'beta', fallback='1.0'))
    buildtime = float(config.get('tweaks', 'packetbuildtime', fallback='0.0000070'))
    accuratets = int(config.get('tweaks', 'accuratetimestamp', fallback='0'))
    print(f"Payload Length= {min} to {max}, Distribution={dist} (alpha={alpha} beta={beta}) Tweak={buildtime}")
    return ConfigSetting(min, max, dist, alpha, beta, buildtime, accuratets)


def create_parser():
    # Argument parser
    parser = argparse.ArgumentParser(description='Send iNetX packets at a specified rate')
    parser.add_argument('--rate', required=False, type=float, default=1.0, help="Packet rate in Mbps")
    parser.add_argument('--ipaddress', required=False, type=str, default="192.168.0.26", help="Destination IP")
    parser.add_argument('--config', required=False, type=str, default="", help="Destination IP")
    parser.add_argument('--maxpps', required=False, type=int, default=None, help="Specify a max pps value")
    parser.add_argument('--datavol', required=False, type=int, default=None, help="Stop after specified bytes")
    parser.add_argument('--sidcount', required=False, type=int, default=1, help="number of stream ids to send")

    return parser


def accurate_sleep(duration, get_now=time.perf_counter):
    now = get_now()
    end = now + duration
    while now < end:
        now = get_now()


def main(args):
    dst_udp_port = 4444
    cfg = get_default_settings(args.config)

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    # Create an inetx packet
    myinetx = inetx.iNetX()

    payload_pkts = {}
    bsid = random.randint(0x0, 0xFF) << 8

    FCS_LEN = 4
    PREAMBLE = 8
    hdr_lens = eth.UDP.UDP_HEADER_SIZE + eth.IP.IP_HEADER_SIZE + eth.Ethernet.HEADERLEN + FCS_LEN + PREAMBLE

    tx_pkt_overhead = cfg.buildtime

    total_vol_data = 0
    timestamp = TS(int(time.time()), 0)
    for idx in range(args.sidcount):
        myinetx.inetxcontrol = inetx.iNetX.DEF_CONTROL_WORD
        myinetx.pif = 0
        myinetx.streamid = bsid + idx
        myinetx.sequence = 0
        if cfg.dist == "beta":
            myinetx.payload = struct.pack(">B", idx+1) * int((cfg.max - cfg.min) * random.betavariate(cfg.alpha, cfg.beta) + cfg.min)
        else:
            myinetx.payload = struct.pack(">B", idx+1) * random.randint(cfg.min, cfg.max)

        myinetx.setPacketTime(timestamp.sec)
        packet_payload = myinetx.pack()
        print("iNetX StreamID={:#0X} Length incl headers={}".format(myinetx.streamid, len(packet_payload) + hdr_lens))
        payload_pkts[myinetx.streamid] = {
            'payload': packet_payload, 'length': len(packet_payload) + eth.UDP.UDP_HEADER_SIZE + eth.IP.IP_HEADER_SIZE
                                                 + eth.Ethernet.HEADERLEN}
        pkt_len = len(packet_payload) + 8 + 20 + 14
        total_vol_data += (pkt_len + hdr_lens)

    chunk_count_ps = args.rate * 1e6/(total_vol_data * 8)
    tx_time_vol = total_vol_data * 8 * 1e-9
    gap_per_pkt = (1 - tx_time_vol) / (args.sidcount * chunk_count_ps) - tx_pkt_overhead
    if gap_per_pkt <= 0:
        gap_per_pkt = 0

    if args.datavol is None:
        expected_end_time = "unknown"
    else:
        expected_end_time = int(args.datavol * 8 / args.rate /60)
    pps = int(args.sidcount * chunk_count_ps)
    if args.maxpps is not None:
        if pps > args.maxpps:
            raise Exception(f"The specified combination has generated a pps of {pps}, creater than {args.maxpps}")
    print("UDP target IP:", args.ipaddress)
    print("UDP target port:", dst_udp_port)
    print("Rate = {} Mbps".format(args.rate))
    print("DLY = {:.6f} s".format(gap_per_pkt))
    print("PPS = {} s".format(pps))

    sequence_roll_over = pow(2, 64)
    pkt_count = {}
    for sid in payload_pkts.keys():
        pkt_count[sid] = 0

    packet_count = 1
    vol_data_sent = 0

    def signal_handler(*args):
        print(f"Exiting. Sent {packet_count:>12,} packets and {vol_data_sent:>15,} bytes")
        sys.exit()

    signal.signal(signal.SIGINT, signal_handler)
    run_sec = 0

    while True:
        random_sid = bsid + (packet_count % args.sidcount)
        random_payload = payload_pkts[random_sid]["payload"]
        if cfg.accuratets:
            mypayload = random_payload[:8] + struct.pack(">I", pkt_count[random_sid]) + \
                        random_payload[12:16] + timestamp.pack() + random_payload[24:]
            timestamp.add(gap_per_pkt)
        else:
            mypayload = random_payload[:8] + struct.pack(">I", pkt_count[random_sid]) + \
                        random_payload[12:]
        # Faster way to build an inetx packet instead of packing the whole header

        pkt_count[random_sid] = (pkt_count[random_sid] + 1) % sequence_roll_over
        sock.sendto(mypayload, (args.ipaddress, dst_udp_port))
        vol_data_sent += (len(mypayload) + hdr_lens)
        accurate_sleep(gap_per_pkt)
        packet_count += 1
        if packet_count % (pps * 30) == 0:
            run_sec += 30
            if args.datavol is not None:
                expected_end_time = int((args.datavol - vol_data_sent) * 8 /(args.rate * 1e6) / 60)
                if vol_data_sent > args.datavol:
                    signal_handler()
            print(f"After {run_sec:>8} seconds : {packet_count:>12,} packets sent. {vol_data_sent:>18,} bytes send."
                  f"End in {expected_end_time} minutes")


if __name__ == '__main__':
    parser = create_parser()
    pargs = parser.parse_args()
    main(pargs)
    sys.exit(0)




from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


class OptionalBuildExt(build_ext):
    """Allow C extension building to fail without aborting the install."""

    def run(self):
        try:
            super().run()
        except Exception:
            print("C extension build failed, using Python fallback.")

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except Exception:
            print(f"Building extension {ext.name} failed, skipping.")


ext_modules = []
try:
    ext_modules.append(
        Extension(
            "AcraNetwork.IRIG106.Chapter7.golay_c",
            ["AcraNetwork/IRIG106/Chapter7/golay_c.c"],
            language="c",
        )
    )
except Exception:
    pass  # Ignore if setup fails to define extension


# Load configuration from setup.cfg
setup(
    ext_modules=ext_modules,
    cmdclass={"build_ext": OptionalBuildExt},
)



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
"""

Record ch10 streams to a ch10 file

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2024"
__version__ = "0.1.0"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import signal
import struct
import AcraNetwork.McastSocket as mcast
import AcraNetwork.SimpleEthernet as SimpleEthernet
import argparse
import sys
import socket
from AcraNetwork.IRIG106.Chapter11 import DataType
import AcraNetwork.IRIG106.Chapter11.ComputerData as chcomputer
import AcraNetwork.IRIG106.Chapter11 as ch10
import logging
from enum import IntEnum

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s %(asctime)s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Format(IntEnum):
    ONE = 1
    TWO = 2
    THREE = 3


class Chapter10FileWriter(object):
    """
    Parse a Chapter10 file. Open the file and iterate through it
    """

    def __init__(self, filename):
        self.filename = filename
        self._mode = "wb"
        self._offset = 0
        self._fd = None

    def write(self, ch10packet: bytes):
        """
        Write a chapter10 packet to the file
        """
        if not self._fd.writable():
            raise Exception("File {} not open for writing".format(self.filename))
        self._fd.write(ch10packet)

    def __enter__(self):
        self._fd = open(self.filename, self._mode)
        return self

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self._fd.close()

    def close(self):
        if self._fd is not None:
            self._fd.close()


def create_parser():
    """Setup a command line parser"""
    description = """Record multicast ch10 packets to a ch10 file. This should be used to capture chapter10 packets. 
    All packets should be setup to transmit using the same format, the same destination multicast IP address and the same destination UDP port"""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--tmats", required=True, help="The TMATs file to place in recorded chapter 10 file")
    parser.add_argument("--ch10", required=True, help="The output chapter 10 filename")
    parser.add_argument("--quiet", required=False, action="store_true", help="Only report error messages")
    parser.add_argument("--udp", required=False, default=51000, type=int, help="The UDP port to capture")
    parser.add_argument(
        "--format",
        required=False,
        default=1,
        choices=[Format.ONE, Format.TWO, Format.THREE],
        type=int,
        help="The chapter 10 format being used. ",
    )
    parser.add_argument("--multicast", required=False, default="235.0.0.1", help="The multicast address to capture")
    return parser


def opensocket(mcast_addr: str, udp_port: int) -> socket.socket:
    """Open a receive UDP socket for packet receiption"""
    try:
        recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        mreq = struct.pack("=4sl", socket.inet_aton(mcast_addr), socket.INADDR_ANY)
        recv_socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        recv_socket.bind(("", udp_port))
        recv_socket.settimeout(1)
    except Exception as e:
        raise Exception(f"Can't bind to socket {udp_port} Err={e}")
    else:
        return recv_socket


def encapsulate_tmats(tmats_filename: str, rtctime: int) -> bytes:
    """Wrap the tmats data in a ch10 file for writing to a ch10 file"""
    with open(tmats_filename, mode="rb") as f:
        tmats = f.read()
        c = ch10.Chapter11()
        c.channelID = 0
        c.sequence = 0
        c.packetflag = 0
        c.datatype = DataType.COMPUTER_FORMAT_1
        c.relativetimecounter = rtctime
        ctmats = chcomputer.ComputerGeneratedFormat1()
        ctmats.payload = tmats
        c.payload = ctmats.pack()
        return c.pack()


def main(args):

    if args.quiet:
        logger.setLevel(logging.ERROR)

    MIN_PKT_SIZE = 30
    # Identify chapter 10 packets
    OFFSET_TO_SYNC = {Format.ONE: 4, Format.TWO: 12, Format.THREE: 8}
    OFFSET_TO_TIME = {Format.ONE: 4 + 15, Format.TWO: 12 + 15, Format.THREE: 8 + 15}

    rec_pkt_count = 0
    keep_recording = True
    try:
        rx_socket = opensocket(args.multicast, args.udp)
    except Exception as e:
        logger.error(f"Failed to capture from network. Error={e}")
        return 1
    else:
        ch10_writer = Chapter10FileWriter(args.ch10)

        def signal_handler(*args):
            logger.info(f"Exiting. Recorded {rec_pkt_count:>12,}")
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)

        time_aligned = False  # Align the writing to the time packet
        with ch10_writer as ch10_file:
            logger.info("Waiting to align to a Chapter 10 time packet")
            while keep_recording:
                try:
                    data, addr = rx_socket.recvfrom(2048)  # buffer size is 1500 bytes
                except Exception as e:
                    logger.error(f"timeout on socket. Err={e}")
                    return 1
                else:
                    if len(data) > MIN_PKT_SIZE:
                        (data_type,) = struct.unpack_from("<B", data, OFFSET_TO_TIME[args.format])
                        logger.debug(f"Rx pkt, dt={data_type:#0X}")
                        if not time_aligned:
                            (data_type, lsw, msw) = struct.unpack_from(
                                "<BIH", data, OFFSET_TO_TIME[args.format]
                            )  # Get the time
                            if 0x11 <= data_type <= 0x17:
                                rtc_count = lsw + (msw << 32)
                                time_aligned = True
                                ch10_file.write(encapsulate_tmats(args.tmats, rtc_count))
                                logger.info("Received time packet. Starting recording")
                        if time_aligned:
                            ch10_payload = data[OFFSET_TO_SYNC[args.format] :]
                            ch10_file.write(ch10_payload)
                            rec_pkt_count += 1
                # Reporting
                if rec_pkt_count % 100 == 0 and rec_pkt_count > 0:
                    logger.info(f"Recorded {rec_pkt_count:>12,} packets")


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    ret = main(args)
    sys.exit(ret)



import typing
import struct
from AcraNetwork.IRIG106.Chapter11 import TS_CH4, TS_IEEE1558, RTCTime, PTPTime


class CANMessage(object):
    """Work in Progress. Do not use

    Args:
        object (_type_): _description_
    """

    def __init__(self, ipts_source=TS_CH4):
        if ipts_source == TS_CH4:
            self.ipts = RTCTime()
        elif ipts_source == TS_IEEE1558:
            self.ipts = PTPTime()
        elif ipts_source is None:
            raise Exception("Time stamp is not option for CAN Bus")

        self.subchannel: int = 0
        self.data_error: bool = False
        self.format_error: bool = False


class CANDataPacket(object):
    def __init__(self, ipts_source=TS_CH4) -> None:
        self.messages: typing.List[CANMessage] = []



__author__ = "diarmuid"
import sys

sys.path.append("..")

import unittest
import AcraNetwork.iNetX as inetx
import AcraNetwork.ParserAligned as paligned
import AcraNetwork.Pcap as pcap
import struct
import os
import copy
from base64 import b64encode

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


expected_p = """Block 0: QuadBytes=3 Error=False ErrorCode=0 BusID=0 MessageCount=0 ElapsedTime=0
Block 1: QuadBytes=3 Error=False ErrorCode=0 BusID=2 MessageCount=1 ElapsedTime=0
Block 2: QuadBytes=3 Error=False ErrorCode=0 BusID=6 MessageCount=2 ElapsedTime=10000
Block 3: QuadBytes=3 Error=False ErrorCode=0 BusID=16 MessageCount=3 ElapsedTime=10000
Block 4: QuadBytes=3 Error=False ErrorCode=0 BusID=18 MessageCount=4 ElapsedTime=20000
Block 5: QuadBytes=3 Error=False ErrorCode=0 BusID=11 MessageCount=5 ElapsedTime=20000
Block 6: QuadBytes=3 Error=False ErrorCode=0 BusID=14 MessageCount=6 ElapsedTime=30000
Block 7: QuadBytes=3 Error=False ErrorCode=0 BusID=17 MessageCount=7 ElapsedTime=30000
Block 8: QuadBytes=3 Error=False ErrorCode=0 BusID=22 MessageCount=8 ElapsedTime=30000
Block 9: QuadBytes=3 Error=False ErrorCode=0 BusID=23 MessageCount=9 ElapsedTime=30000
Block 10: QuadBytes=3 Error=False ErrorCode=0 BusID=7 MessageCount=10 ElapsedTime=30000
Block 11: QuadBytes=3 Error=False ErrorCode=0 BusID=8 MessageCount=11 ElapsedTime=30000
Block 12: QuadBytes=3 Error=False ErrorCode=0 BusID=9 MessageCount=12 ElapsedTime=40000
Block 13: QuadBytes=3 Error=False ErrorCode=0 BusID=10 MessageCount=13 ElapsedTime=50000
Block 14: QuadBytes=3 Error=False ErrorCode=0 BusID=12 MessageCount=14 ElapsedTime=50000
Block 15: QuadBytes=3 Error=False ErrorCode=0 BusID=13 MessageCount=15 ElapsedTime=50000
Block 16: QuadBytes=3 Error=False ErrorCode=0 BusID=15 MessageCount=16 ElapsedTime=40000
Block 17: QuadBytes=3 Error=False ErrorCode=0 BusID=19 MessageCount=17 ElapsedTime=50000
Block 18: QuadBytes=3 Error=False ErrorCode=0 BusID=20 MessageCount=18 ElapsedTime=50000
Block 19: QuadBytes=3 Error=False ErrorCode=0 BusID=21 MessageCount=19 ElapsedTime=40000
Block 20: QuadBytes=3 Error=False ErrorCode=0 BusID=1 MessageCount=20 ElapsedTime=50000
Block 21: QuadBytes=3 Error=False ErrorCode=0 BusID=3 MessageCount=21 ElapsedTime=50000
Block 22: QuadBytes=3 Error=False ErrorCode=0 BusID=4 MessageCount=22 ElapsedTime=50000
Block 23: QuadBytes=3 Error=False ErrorCode=0 BusID=5 MessageCount=23 ElapsedTime=50000
Block 24: QuadBytes=3 Error=False ErrorCode=0 BusID=2 MessageCount=24 ElapsedTime=360000
Block 25: QuadBytes=3 Error=False ErrorCode=0 BusID=11 MessageCount=25 ElapsedTime=370000
Block 26: QuadBytes=3 Error=False ErrorCode=0 BusID=16 MessageCount=26 ElapsedTime=370000
Block 27: QuadBytes=3 Error=False ErrorCode=0 BusID=0 MessageCount=27 ElapsedTime=370000
Block 28: QuadBytes=3 Error=False ErrorCode=0 BusID=23 MessageCount=28 ElapsedTime=380000
Block 29: QuadBytes=3 Error=False ErrorCode=0 BusID=7 MessageCount=29 ElapsedTime=390000
Block 30: QuadBytes=3 Error=False ErrorCode=0 BusID=15 MessageCount=30 ElapsedTime=410000
Block 31: QuadBytes=3 Error=False ErrorCode=0 BusID=17 MessageCount=31 ElapsedTime=410000
Block 32: QuadBytes=3 Error=False ErrorCode=0 BusID=18 MessageCount=32 ElapsedTime=410000
Block 33: QuadBytes=3 Error=False ErrorCode=0 BusID=21 MessageCount=33 ElapsedTime=410000
Block 34: QuadBytes=3 Error=False ErrorCode=0 BusID=22 MessageCount=34 ElapsedTime=420000
Block 35: QuadBytes=3 Error=False ErrorCode=0 BusID=3 MessageCount=35 ElapsedTime=410000
Block 36: QuadBytes=3 Error=False ErrorCode=0 BusID=4 MessageCount=36 ElapsedTime=420000
Block 37: QuadBytes=3 Error=False ErrorCode=0 BusID=5 MessageCount=37 ElapsedTime=420000
Block 38: QuadBytes=3 Error=False ErrorCode=0 BusID=6 MessageCount=38 ElapsedTime=410000
Block 39: QuadBytes=3 Error=False ErrorCode=0 BusID=8 MessageCount=39 ElapsedTime=420000
Block 40: QuadBytes=3 Error=False ErrorCode=0 BusID=9 MessageCount=40 ElapsedTime=420000
Block 41: QuadBytes=3 Error=False ErrorCode=0 BusID=10 MessageCount=41 ElapsedTime=430000
Block 42: QuadBytes=3 Error=False ErrorCode=0 BusID=12 MessageCount=42 ElapsedTime=420000
Block 43: QuadBytes=3 Error=False ErrorCode=0 BusID=13 MessageCount=43 ElapsedTime=440000
Block 44: QuadBytes=3 Error=False ErrorCode=0 BusID=14 MessageCount=44 ElapsedTime=420000
Block 45: QuadBytes=3 Error=False ErrorCode=0 BusID=19 MessageCount=45 ElapsedTime=440000
Block 46: QuadBytes=3 Error=False ErrorCode=0 BusID=20 MessageCount=46 ElapsedTime=430000
Block 47: QuadBytes=3 Error=False ErrorCode=0 BusID=1 MessageCount=47 ElapsedTime=430000
Block 48: QuadBytes=3 Error=False ErrorCode=0 BusID=16 MessageCount=48 ElapsedTime=720000
Block 49: QuadBytes=3 Error=False ErrorCode=0 BusID=2 MessageCount=49 ElapsedTime=750000
Block 50: QuadBytes=3 Error=False ErrorCode=0 BusID=7 MessageCount=50 ElapsedTime=750000
Block 51: QuadBytes=3 Error=False ErrorCode=0 BusID=11 MessageCount=51 ElapsedTime=760000
Block 52: QuadBytes=3 Error=False ErrorCode=0 BusID=15 MessageCount=52 ElapsedTime=760000
Block 53: QuadBytes=3 Error=False ErrorCode=0 BusID=0 MessageCount=53 ElapsedTime=760000
Block 54: QuadBytes=3 Error=False ErrorCode=0 BusID=4 MessageCount=54 ElapsedTime=770000
Block 55: QuadBytes=3 Error=False ErrorCode=0 BusID=9 MessageCount=55 ElapsedTime=770000
Block 56: QuadBytes=3 Error=False ErrorCode=0 BusID=17 MessageCount=56 ElapsedTime=770000
Block 57: QuadBytes=3 Error=False ErrorCode=0 BusID=18 MessageCount=57 ElapsedTime=770000
Block 58: QuadBytes=3 Error=False ErrorCode=0 BusID=23 MessageCount=58 ElapsedTime=770000
Block 59: QuadBytes=3 Error=False ErrorCode=0 BusID=6 MessageCount=59 ElapsedTime=780000
Block 60: QuadBytes=3 Error=False ErrorCode=0 BusID=8 MessageCount=60 ElapsedTime=780000
Block 61: QuadBytes=3 Error=False ErrorCode=0 BusID=13 MessageCount=61 ElapsedTime=790000
Block 62: QuadBytes=3 Error=False ErrorCode=0 BusID=20 MessageCount=62 ElapsedTime=790000
Block 63: QuadBytes=3 Error=False ErrorCode=0 BusID=21 MessageCount=63 ElapsedTime=790000
Block 64: QuadBytes=3 Error=False ErrorCode=0 BusID=22 MessageCount=64 ElapsedTime=800000
Block 65: QuadBytes=3 Error=False ErrorCode=0 BusID=3 MessageCount=65 ElapsedTime=800000
Block 66: QuadBytes=3 Error=False ErrorCode=0 BusID=10 MessageCount=66 ElapsedTime=800000
Block 67: QuadBytes=3 Error=False ErrorCode=0 BusID=12 MessageCount=67 ElapsedTime=800000
Block 68: QuadBytes=3 Error=False ErrorCode=0 BusID=14 MessageCount=68 ElapsedTime=810000
Block 69: QuadBytes=3 Error=False ErrorCode=0 BusID=1 MessageCount=69 ElapsedTime=820000
Block 70: QuadBytes=3 Error=False ErrorCode=0 BusID=5 MessageCount=70 ElapsedTime=820000
Block 71: QuadBytes=3 Error=False ErrorCode=0 BusID=19 MessageCount=71 ElapsedTime=830000
Block 72: QuadBytes=3 Error=False ErrorCode=0 BusID=16 MessageCount=72 ElapsedTime=1080000
Block 73: QuadBytes=3 Error=False ErrorCode=0 BusID=2 MessageCount=73 ElapsedTime=1110000
Block 74: QuadBytes=3 Error=False ErrorCode=0 BusID=17 MessageCount=74 ElapsedTime=1120000
Block 75: QuadBytes=3 Error=False ErrorCode=0 BusID=18 MessageCount=75 ElapsedTime=1130000
Block 76: QuadBytes=3 Error=False ErrorCode=0 BusID=23 MessageCount=76 ElapsedTime=1130000
Block 77: QuadBytes=3 Error=False ErrorCode=0 BusID=4 MessageCount=77 ElapsedTime=1130000
Block 78: QuadBytes=3 Error=False ErrorCode=0 BusID=6 MessageCount=78 ElapsedTime=1130000
Block 79: QuadBytes=3 Error=False ErrorCode=0 BusID=7 MessageCount=79 ElapsedTime=1140000
Block 80: QuadBytes=3 Error=False ErrorCode=0 BusID=0 MessageCount=80 ElapsedTime=1140000
Block 81: QuadBytes=3 Error=False ErrorCode=0 BusID=9 MessageCount=81 ElapsedTime=1150000
Block 82: QuadBytes=3 Error=False ErrorCode=0 BusID=12 MessageCount=82 ElapsedTime=1150000
"""


class TestParserAligned(unittest.TestCase):
    def test_read_pcap(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "valid_paligned.pcap"))
        mypcaprecord = p[0]
        p.close()
        # Now I have a _payload that will be an inetx packet
        inetxp = inetx.iNetX()
        inetxp.unpack(mypcaprecord.payload[0x2A:-4])
        self.assertEqual(inetxp.streamid, 0xA00)
        p = paligned.ParserAlignedPacket()
        p.unpack(inetxp.payload)
        self.assertEqual(len(p), 83)
        self.assertEqual(repr(p), expected_p)
        for i, b in enumerate(p):
            if i == 3:
                self.assertEqual(len(b), 12)
                self.assertEqual(
                    repr(b), "QuadBytes=3 Error=False ErrorCode=0 BusID=16 MessageCount=3 ElapsedTime=10000"
                )

        p2 = paligned.ParserAlignedPacket()
        for b in p:
            blk = copy.copy(b)
            p2.parserblocks.append(blk)

        self.assertEqual(p2.pack(), inetxp.payload)
        self.assertTrue(p2 == p)
        p3 = paligned.ParserAlignedPacket()
        for idx, b in enumerate(p):
            blk = copy.copy(b)
            p3.parserblocks.append(blk)
            if idx == 2:
                break
        i = inetx.iNetX()
        i.streamid = 1
        i.payload = p3.pack()
        print(b64encode(i.pack()))


if __name__ == "__main__":
    unittest.main()



__author__ = "diarmuid"
import sys

sys.path.append("..")

import unittest
import AcraNetwork.iNetX as inetx
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct
import os
import copy
from base64 import b64encode


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class iNetXTest(unittest.TestCase):
    def test_defaultiNet(self):
        i = inetx.iNetX()
        self.assertEqual(i.packetlen, 0)
        self.assertEqual(i.inetxcontrol, inetx.iNetX.DEF_CONTROL_WORD)
        self.assertEqual(i.ptptimeseconds, 0)
        self.assertEqual(i.ptptimenanoseconds, 0)
        self.assertEqual(i.pif, 0)
        self.assertEqual(i.payload, bytes())
        self.assertEqual(i.sequence, 0)
        self.assertEqual(i.streamid, 0)

    def test_basiciNet(self):
        i = inetx.iNetX()
        i.sequence = 1
        i.pif = 0
        i.streamid = 0xDC
        i.setPacketTime(1, 1)
        i.payload = struct.pack("H", 0x5)
        expected_payload = struct.pack(
            inetx.iNetX.INETX_HEADER_FORMAT, inetx.iNetX.DEF_CONTROL_WORD, 0xDC, 1, 30, 1, 1, 0
        ) + struct.pack("H", 0x5)
        print(b64encode(expected_payload))
        self.assertEqual(i.pack(), expected_payload)

    def test_unpackiNet(self):
        expected_payload = struct.pack(
            inetx.iNetX.INETX_HEADER_FORMAT, inetx.iNetX.DEF_CONTROL_WORD, 0xDC, 2, 30, 1, 1, 0
        ) + struct.pack("H", 0x5)
        i = inetx.iNetX(expected_payload)
        self.assertEqual(i.sequence, 2)
        self.assertEqual(i.streamid, 0xDC)
        self.assertEqual(i.ptptimeseconds, 1)
        self.assertEqual(i.ptptimenanoseconds, 1)
        self.assertEqual(i.packetlen, 30)
        self.assertEqual(i.pif, 0)
        self.assertEqual(i.payload, struct.pack("H", 0x5))
        self.assertEqual(repr(i), "STREAMID=0XDC SEQ=2 LEN=30 PTPS=1 PTPNS=1")
        i2 = copy.copy(i)
        self.assertEqual(i, i2)
        self.assertEqual(len(i), 28 + 2)

    def test_unpackiNetFromPcap(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "inetx_test.pcap"))
        mypcaprecord = p[0]
        p.close()
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip = SimpleEthernet.IP()
        ip.unpack(e.payload)
        u = SimpleEthernet.UDP()
        u.unpack(ip.payload)
        # Now I have a _payload that will be an inetx packet
        i = inetx.iNetX()
        i.unpack(u.payload)
        self.assertEqual(i.inetxcontrol, 0x11000000)
        self.assertEqual(i.streamid, 0xCA)
        self.assertEqual(i.sequence, 1011)
        self.assertEqual(i.packetlen, 72)
        self.assertEqual(i.ptptimeseconds, 0x2F3)
        self.assertEqual(i.ptptimenanoseconds, 0x2CB4158C)
        self.assertEqual(i.pif, 0)

    def test_unpackMultiplePackets(self):
        sequencenum = 1011
        mypcap = pcap.Pcap(os.path.join(THIS_DIR, "inetx_test.pcap"))  # Read the pcap file
        for mypcaprecord in mypcap:
            ethpacket = SimpleEthernet.Ethernet()  # Create an Ethernet object
            ethpacket.unpack(mypcaprecord.packet)  # Unpack the pcap record into the eth object
            ippacket = SimpleEthernet.IP()  # Create an IP packet
            ippacket.unpack(ethpacket.payload)  # Unpack the ethernet _payload into the IP packet
            udppacket = SimpleEthernet.UDP()  # Create a UDP packet
            udppacket.unpack(ippacket.payload)  # Unpack the IP _payload into the UDP packet
            inetxpacket = inetx.iNetX()  # Create an iNetx object
            inetxpacket.unpack(udppacket.payload)  # Unpack the UDP _payload into this iNetX object
            # print "INETX: StreamID ={:08X} Sequence = {:8d} PTP Seconds = {}".format(inetxpacket.streamid,inetxpacket.sequence,inetxpacket.ptptimeseconds)
            self.assertEqual(inetxpacket.sequence, sequencenum)
            sequencenum += 1
        mypcap.close()


if __name__ == "__main__":
    unittest.main()



import struct
from AcraNetwork.IRIG106.Chapter11 import TS_CH4, TS_IEEE1558, RTCTime, PTPTime
from enum import IntEnum, unique
import typing


@unique
class Endianness(IntEnum):
    BIG: int = 0
    LITTLE: int = 1


def endian_swap(buffer: bytes) -> bytes:
    return bytearray([c for t in zip(buffer[1::2], buffer[::2]) for c in t])


class UARTDataWord(object):
    """
    The Chapter 10 standard defines specific payload formats for different data. This class handles UART packets

    :type ptptimeseconds: int
    :type ptptimenanoseconds: int
    :type subchannel: int
    :type parity_error: bool
    :type datalength: int
    :type payload: str
    """

    def __init__(self, ipts_source=TS_CH4, data_endianness: int = Endianness.BIG):
        if ipts_source == TS_CH4:
            self.ipts = RTCTime()
        elif ipts_source == TS_IEEE1558:
            self.ipts = PTPTime()
        elif ipts_source is None:
            self.ipts = None
        self.parity_error: bool = False  #: Parity error has occurred
        self.subchannel: int = 0  #: Subchannel
        self.datalength: int = None  #: Data Length
        self._payload: bytes = bytes()  #: UART payload
        self.data_endianness: int = data_endianness

    @property
    def payload(self) -> bytes:
        return self._payload

    @payload.setter
    def payload(self, mybuffer: bytes):
        self._payload = mybuffer
        self.datalength = len(mybuffer)

    def pack(self) -> bytes:
        """
        Pack the UART data packet object into a binary buffer

        :rtype: str|bytes
        """
        if self.ipts is None:
            ch_spec_word = b""
        else:
            ch_spec_word = self.ipts.pack()
        data_len = len(self.payload)
        if self.parity_error:
            _subch = self.subchannel + 0x80
        else:
            _subch = self.subchannel
        intra_pkt_header = struct.pack("<HH", data_len, _subch)

        if data_len % 2 == 1:
            padding = struct.pack("B", 0xFF)
        else:
            padding = b""

        payload_and_padding = self.payload + padding
        if self.data_endianness == Endianness.LITTLE:
            payload_and_padding = endian_swap(payload_and_padding)
        return ch_spec_word + intra_pkt_header + payload_and_padding

    def unpack(self, mybuffer: bytes):
        """
        Unpack a string buffer into an UART data packet object. Returns the buffer that was consumed

        :param mybuffer: A string buffer representing an UART data  packet
        :type mybuffer: str
        :param timestamp_present: The UART DataPacket Format knows if there is a timestamp or not
        :type timestamp_present: bool
        :rtype: int
        """
        offset = 0
        if self.ipts is not None:
            # bytes = struct.unpack_from(">8B", mybuffer)
            self.ipts.unpack(mybuffer[:8])
            offset += 8
        (self.datalength, _pe_sub) = struct.unpack_from("<HH", mybuffer, offset)
        offset += 4

        self.subchannel = _pe_sub & 0x1FFF
        self.parity_error = bool(_pe_sub >> 15)

        if self.data_endianness == Endianness.LITTLE:
            if self.datalength % 2 == 1:
                payload_and_padding = mybuffer[offset : offset + self.datalength + 1]  # take the buffer and the payload
                payload_and_padding = endian_swap(payload_and_padding)
                payload_and_padding = payload_and_padding[:-1]
            else:
                payload_and_padding = mybuffer[offset : offset + self.datalength]
                payload_and_padding = endian_swap(payload_and_padding)
        else:
            payload_and_padding = mybuffer[offset : offset + self.datalength]

        self.payload = payload_and_padding
        offset += self.datalength

        if self.datalength % 2 == 1:
            return offset + 1
        else:
            return offset

    def __eq__(self, other):
        if not isinstance(other, UARTDataWord):
            return False

        _match_att = ["ipts", "parity_error", "subchannel", "datalength", "payload"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return f"UARTDataWord: Time={self.ipts} ParityError={self.parity_error} DataLen={self.datalength} SubChannel={self.subchannel} Endianness={repr(self.data_endianness)}"


class UARTDataPacket(object):
    """
    Data Packet Format. Contains a list of UART Data Words

    :type msgcount: int
    :type uartwords: list[UARTDataWord]


    >>> import AcraNetwork.Pcap as pcap
    >>> from AcraNetwork.IRIG106.Chapter11 import Chapter11
    >>> p = pcap.Pcap("test/sample_pcap/sample_ch10uart.pcap")
    >>> data = p[0].payload
    >>> c = Chapter11()
    >>> c.unpack(data[0x2E:])
    True
    >>> uart_p = UARTDataPacket()
    >>> uart_p.unpack(c.payload)
    True
    >>> print(len(uart_p))
    22
    >>> print(uart_p[0])
    UARTDataWord: Time=RTC: count=43151323112 ParityError=True DataLen=34 SubChannel=0 Endianness=<Endianness.BIG: 0>

    """

    def __init__(self, ipts_source=TS_CH4, data_endianness: int = Endianness.BIG):
        self.uartwords: typing.List[UARTDataWord] = []  #: List of :class:`UARTDataWord`
        self._ipts_source = ipts_source
        self.data_endianness: int = data_endianness

    def pack(self) -> bytes:
        """
        Pack the UART data packet object into a binary buffer

        :rtype: str
        """
        # Some checks
        if len(self) == 0:
            raise Exception("No UARTDataWords defined")

        # All words should have the same configuraiton of timestamps
        if self._ipts_source is not None:
            ret_buf = struct.pack("<I", 0x1 << 31)
        else:
            ret_buf = struct.pack("<I", 0x0 << 31)

        for udw in self:
            ret_buf += udw.pack()

        return ret_buf

    def unpack(self, mybuffer: bytes):
        """
        Unpack a string buffer into an UART data packet object

        :param buffer: A string buffer representing an UART data  packet
        :type buffer: str
        :rtype: None
        """

        (ch_spec_word,) = struct.unpack_from("<I", mybuffer)
        ts_present = bool(ch_spec_word >> 31)
        offset = 4
        while abs(offset - len(mybuffer)) > 4:
            udw = UARTDataWord(self._ipts_source, self.data_endianness)
            offset += udw.unpack(mybuffer[offset:])
            self.uartwords.append(udw)

        return True

    def append(self, udw: UARTDataWord):
        """
        Add a UART DW to the DP

        :type udw: UARTDataWord
        :rtype: bool
        """
        if not isinstance(udw, UARTDataWord):
            raise Exception("Can only append UARTDataWords")
        return self.uartwords.append(udw)

    def __eq__(self, other: UARTDataWord):
        if not isinstance(other, UARTDataPacket):
            return False

        _match_att = ["uartwords"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        ret_str = "UARTPayload: UARTDataWordCount={}\n".format(len(self))

        for a in self:
            ret_str += "  {}\n".format(repr(a))

        return ret_str

    def __iter__(self):
        self._index = 0
        return self

    def next(self) -> UARTDataWord:
        if self._index < len(self.uartwords):
            _dw = self.uartwords[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.uartwords)

    def __getitem__(self, key):
        return self.uartwords[key]
